import re
import collections


class BytePairEncoder:
    def __init__(self, whitespace_token="▁", unknown_token="</u>", lowercase=True):
        self.merges = {}
        self.tokens = {}
        self.sorted_tokens = []
        self.vocab = {}
        self.vocab_tokenized = {}
        self.whitespace_token = whitespace_token
        self.lowercase = lowercase
        self.unknown_token = unknown_token

    def format_word(self, word):
        return " ".join(list(word))

    def initialize_vocab(self, texts):
        vocab = {}
        for text in texts:
            words = self.pre_tokenize(text)
            for word in words:
                word = self.format_word(word)
                vocab[word] = vocab.get(word, 0) + 1
        return vocab

    def get_tokens_from_vocab(self, vocab):
        tokens = collections.defaultdict(int)
        vocab_tokenized = {}
        for word, freq in vocab.items():
            word_tokens = word.split()
            for token in word_tokens:
                tokens[token] += freq
            vocab_tokenized["".join(word_tokens)] = word_tokens
        return tokens, vocab_tokenized

    def get_bigram_counts(self, vocab):
        pairs = {}
        for word, count in vocab.items():
            symbols = word.split()
            for i in range(len(symbols) - 1):
                pair = (symbols[i], symbols[i + 1])
                pairs[pair] = pairs.get(pair, 0) + count
        return pairs

    def merge_vocab(self, pair, vocab_in):
        vocab_out = {}
        bigram = re.escape(" ".join(pair))
        p = re.compile(r"(?<!\S)" + bigram + r"(?!\S)")
        bytepair = "".join(pair)
        for word in vocab_in:
            w_out = p.sub(bytepair, word)
            vocab_out[w_out] = vocab_in[word]
        return vocab_out, (bigram, bytepair)

    def find_merges(self, vocab, tokens, num_merges, indices_to_print=[]):
        merges = []
        for i in range(num_merges):
            pairs = self.get_bigram_counts(vocab)
            best_pair = max(pairs, key=pairs.get)
            best_count = pairs[best_pair]
            vocab, (bigram, bytepair) = self.merge_vocab(best_pair, vocab)
            merges.append((r"(?<!\S)" + bigram + r"(?!\S)", bytepair))
            tokens, vocab_tokenized = self.get_tokens_from_vocab(vocab)
            if indices_to_print and i in indices_to_print:
                print(f"Merge {i}: {best_pair} with count {best_count}")
                print("All tokens: {}".format(tokens.keys()))
                print("Number of tokens: {}".format(len(tokens.keys())))
        return vocab, tokens, merges, vocab_tokenized

    def fit(self, texts, num_merges, indices_to_print=[]):
        vocab = self.initialize_vocab(texts)

        self.vocab, self.tokens, self.merges, self.vocab_tokenized = self.find_merges(
            vocab, self.tokens, num_merges, indices_to_print
        )
        self.sorted_tokens = sorted(
            self.tokens.keys(),
            key=self.measure_token_length,
            reverse=True,
        )

    def measure_token_length(self, token):
        whitespace_symbol_len = len(self.whitespace_token)
        if token[-whitespace_symbol_len:] == self.whitespace_token:
            return len(token) - whitespace_symbol_len + 1
        else:
            return len(token)

    def encode_word(self, string, sorted_tokens):

        if string == "":
            return []
        sorted_tokens = sorted_tokens.copy()
        if sorted_tokens == []:
            return [self.unknown_token]

        string_tokens = []
        for i in range(len(sorted_tokens)):
            token = sorted_tokens[i]
            token_reg = re.escape(token.replace(".", "[.]"))

            matched_positions = [
                (m.start(0), m.end(0)) for m in re.finditer(token_reg, string)
            ]
            if len(matched_positions) == 0:
                continue
            substring_end_positions = [
                matched_position[0] for matched_position in matched_positions
            ]

            substring_start_position = 0
            for substring_end_position in substring_end_positions:
                substring = string[substring_start_position:substring_end_position]
                string_tokens += self.encode_word(
                    substring, sorted_tokens=sorted_tokens[i + 1 :]
                )
                string_tokens += [token]
                substring_start_position = substring_end_position + len(token)
            remaining_substring = string[substring_start_position:]
            string_tokens += self.encode_word(
                remaining_substring, sorted_tokens=sorted_tokens[i + 1 :]
            )
            break
        return string_tokens

    def print_enocoded_word(self, word):
        print("Encoding word: {}...".format(word))
        if word in self.vocab_tokenized:
            print("Encoding of the known word:")
            print(self.vocab_tokenized[word])
            print("Encoding treating the known word as unknown:")
            print(self.encode_word(word, self.sorted_tokens))
        else:
            print("Encoding of the unknown word:")
            print(self.encode_word(word, self.sorted_tokens))

    def pre_tokenize(self, text):
        if self.lowercase:
            text = text.lower()
        text = re.sub(r"\s+", " ", text)
        return [word + self.whitespace_token for word in text.split()]

    def tokenize(self, text):
        words = self.pre_tokenize(text)
        encoded_words = [self.encode_word(word, self.sorted_tokens) for word in words]
        return sum(encoded_words, [])



from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import classification_report
from sklearn.metrics import accuracy_score
from sklearn.metrics import classification_report


def evaluate_classification_performance(
    true_labels, predicted_labels, average="weighted", labels=None, **kwargs
):
    print("Accuracy: ", accuracy_score(true_labels, predicted_labels))
    print("Precison: ", precision_score(true_labels, predicted_labels, average=average))
    print("Recall: ", recall_score(true_labels, predicted_labels, average=average))
    print("F1 Score: ", f1_score(true_labels, predicted_labels, average=average))
    print(
        "Model Report: \n___________________________________________________",
    )
    print(classification_report(true_labels, predicted_labels))
    cm = confusion_matrix(true_labels, predicted_labels, labels=labels)

    return cm



import ast
import time
import os
import re
import datetime
import chardet
from contextlib import contextmanager
from timeit import default_timer


def unescape_dict(d):
    return ast.literal_eval(repr(d).encode("utf-8").decode("unicode-escape"))


@contextmanager
def elapsed_timer(format_time=False):
    start = default_timer()
    elapser = lambda: default_timer() - start
    yield lambda: str(
        datetime.timedelta(seconds=elapser())
    ) if format_time else elapser()
    end = default_timer()
    elapser = lambda: end - start


def lower_case_with_underscores(s):
    return re.sub(r"\s+", "_", s.lower()).replace("-", "_")


ordinal = lambda n: "%d%s" % (
    n,
    "tsnrhtdd"[(n // 10 % 10 != 1) * (n % 10 < 4) * n % 10 :: 4],
)


def get_offset_ranges(count, num_workers):
    assert count > num_workers
    step_sz = int(count / num_workers)
    offset_ranges = [0]
    pv_cnt = 1
    for i in range(num_workers):
        if i == num_workers - 1:
            pv_cnt = count
        else:
            pv_cnt = pv_cnt + step_sz
        offset_ranges.append(pv_cnt)
    return offset_ranges


@contextmanager
def change_directory(directory):
    original = os.path.abspath(os.getcwd())

    fancy_print(" Change directory to {}".format(directory))
    os.chdir(directory)
    yield

    fancy_print(" Change directory back to {}".format(original))
    os.chdir(original)


def fancy_print(*args, color=None, bold=False, **kwargs):
    if bold:
        print("\033[1m", end="")

    if color:
        print("\033[{}m".format(color), end="")

    print(*args, **kwargs)

    print("\033[0m", end="")  # reset


# https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb/37423778
def humanbytes(B, units=None):
    "Return the given bytes as a human friendly KB, MB, GB, or TB string"
    B = float(B)
    KB = float(1024)
    MB = float(KB ** 2)  # 1,048,576
    GB = float(KB ** 3)  # 1,073,741,824
    TB = float(KB ** 4)  # 1,099,511,627,776

    if (B < KB and units is None) or units == "B":
        return "{0} {1}".format(B, "Bytes" if 0 == B > 1 else "Byte")
    elif (KB <= B < MB and units is None) or units == "KiB":
        return "{0:.2f} KiB".format(B / KB)
    elif (MB <= B < GB and units is None) or units == "MiB":
        return "{0:.2f} MiB".format(B / MB)
    elif (GB <= B < TB and units is None) or units == "GiB":
        return "{0:.2f} GiB".format(B / GB)
    elif (TB <= B and units is None) or units == "TiB":
        return "{0:.2f} TiB".format(B / TB)


def parse_size(sizestr):
    unit = sizestr[-1]
    size = float(sizestr[:-1])

    if unit.upper() == "B":
        return size
    if unit.upper() == "K":
        return size * 1024
    if unit.upper() == "M":
        return size * 1024 * 1024
    if unit.upper() == "G":
        return size * 1024 * 1024 * 1024
    if unit.upper() == "T":
        return size * 1024 * 1024 * 1024 * 1024


def check_min_len(s, len_func, min_len):
    return len_func(s) >= min_len


def check_max_len(s, len_func, max_len):
    return len_func(s) <= max_len


def utf8len(s):
    return len(str(s).encode("utf-8"))


def len_wospc(x):
    return utf8len(re.sub(r"\s", "", str(x)))


def len_bytes(x):
    return utf8len(x)


def len_words(x):
    if isinstance(x, str):
        return len(x.split())
    return 0


def len_sents(x, sep):
    sep = str(sep).encode("utf-8").decode("unicode-escape")
    return len(re.sub(r"(\r?\n|\r){1,}", sep, x).split(sep))


def len_segments(x, sep):
    sep = str(sep).encode("utf-8").decode("unicode-escape")
    return len(re.sub(r"(\r?\n|\r){2,}", sep, x).split(sep))


def any_to_utf8(b):
    try:
        return b.decode("utf-8")
    except UnicodeDecodeError:
        # try to figure out encoding if not urf-8

        guess = chardet.detect(b)["encoding"]

        if not guess or guess == "UTF-8":
            return

        try:
            return b.decode(guess)
        except (UnicodeDecodeError, LookupError):
            # still cant figure out encoding, give up
            return


def get_modified_time(path):
    if not os.path.exists(path):
        return None
    modTimesinceEpoc = os.path.getmtime(path)
    # Convert seconds since epoch to readable timestamp
    modificationTime = time.strftime(
        "%Y-%m-%d %H:%M:%S", time.localtime(modTimesinceEpoc)
    )
    return modificationTime



"""PyTorch-based multi-task learning framework for discriminative modeling."""

from .data import DictDataLoader, DictDataset  # noqa: F401
from .loss import cross_entropy_with_probs  # noqa: F401
from .multitask_classifier import MultitaskClassifier  # noqa: F401
from .task import Operation, Task  # noqa: F401
from .training.loggers import (  # noqa: F401
    Checkpointer,
    CheckpointerConfig,
    LogManager,
    LogManagerConfig,
    LogWriter,
    LogWriterConfig,
    TensorBoardWriter,
)
from .training.trainer import Trainer  # noqa: F401



from typing import List, Optional, Sequence

import numpy as np

from .core import Policy


class MeanFieldPolicy(Policy):
    """Sample sequences of TFs according to a distribution.

    Samples sequences of indices of a specified length from a
    user-provided distribution. A distribution over TFs can be
    learned by a TANDA mean-field model, for example.
    See https://hazyresearch.github.io/snorkel/blog/tanda.html

    Parameters
    ----------
    n_tfs
        Total number of TFs
    sequence_length
        Number of TFs to run on each data point
    p
        Probability distribution from which to sample TF indices.
        Must have length ``n_tfs`` and be a valid distribution.
    n_per_original
        Number of transformed data points per original
    keep_original
        Keep untransformed data point in augmented data set? Note that
        even if in-place modifications are made to the original data
        point by the TFs being applied, the original data point will
        remain unchanged.

    Attributes
    ----------
    n
        Total number of TFs
    n_per_original
        See above
    keep_original
        See above
    sequence_length
        See above
    """

    def __init__(
        self,
        n_tfs: int,
        sequence_length: int = 1,
        p: Optional[Sequence[float]] = None,
        n_per_original: int = 1,
        keep_original: bool = True,
    ) -> None:
        self.sequence_length = sequence_length
        self._p = p
        super().__init__(
            n_tfs, n_per_original=n_per_original, keep_original=keep_original
        )

    def generate(self) -> List[int]:
        """Generate a sequence of TF indices by sampling from distribution.

        Returns
        -------
        List[int]
            Indices of TFs to run on data point in order.
        """
        return np.random.choice(self.n, size=self.sequence_length, p=self._p).tolist()


class RandomPolicy(MeanFieldPolicy):
    """Naive random augmentation policy.

    Samples sequences of TF indices a specified length at random
    from the total number of TFs. Sampling uniformly at random is
    a common baseline approach to data augmentation.

    Parameters
    ----------
    n_tfs
        Total number of TFs
    sequence_length
        Number of TFs to run on each data point
    n_per_original
        Number of transformed data points per original
    keep_original
        Keep untransformed data point in augmented data set? Note that
        even if in-place modifications are made to the original data
        point by the TFs being applied, the original data point will
        remain unchanged.

    Attributes
    ----------
    n
        Total number of TFs
    n_per_original
        See above
    keep_original
        See above
    sequence_length
        See above
    """

    def __init__(
        self,
        n_tfs: int,
        sequence_length: int = 1,
        n_per_original: int = 1,
        keep_original: bool = True,
    ) -> None:
        super().__init__(
            n_tfs,
            sequence_length=sequence_length,
            p=None,
            n_per_original=n_per_original,
            keep_original=keep_original,
        )



from typing import Set, Type

from .http import HttpClient
from .scheme_client import SchemeClient
from .hf import hf_get_from_cache

# from .s3 import S3Client
# from .gs import GsClient

_SCHEME_TO_CLIENT = {}


def add_scheme_client(client: Type[SchemeClient]) -> None:
    """
    Add a new :class:`SchemeClient`.

    This can be used to extend :func:`cached_path.cached_path()` to handle custom schemes, or handle
    existing schemes differently.
    """
    global _SCHEME_TO_CLIENT
    if isinstance(client.scheme, tuple):
        for scheme in client.scheme:
            _SCHEME_TO_CLIENT[scheme] = client
    elif isinstance(client.scheme, str):
        _SCHEME_TO_CLIENT[client.scheme] = client
    else:
        raise ValueError(f"Unexpected type for {client} scheme: {client.scheme}")


# for client in (HttpClient, S3Client, GsClient):
for client in [HttpClient]:
    add_scheme_client(client)  # type: ignore


def get_scheme_client(resource: str) -> SchemeClient:
    """
    Get the right client for the given resource.
    """
    maybe_scheme = resource.split("://")[0]
    return _SCHEME_TO_CLIENT.get(maybe_scheme, HttpClient)(resource)


def get_supported_schemes() -> Set[str]:
    """
    Return all supported URL schemes.
    """
    return set(_SCHEME_TO_CLIENT.keys()) | {"hf"}



from typing import Dict, List

import numpy as np
import torch
from numpy.lib import recfunctions as rfn
from torch import nn

from ..analysis import Scorer
from ..classification.data import DictDataLoader
from ..classification.multitask_classifier import Operation, Task

from .modules.slice_combiner import SliceCombinerModule


def add_slice_labels(
    dataloader: DictDataLoader, base_task: Task, S: np.recarray
) -> None:
    """Modify a dataloader in-place, adding labels for slice tasks.

    Parameters
    ----------
    dataloader
        A DictDataLoader whose dataset.Y_dict attribute will be modified in place
    base_task
       The Task for which we want corresponding slice tasks/labels
    S
        A recarray (output of SFApplier) containing data fields with slice
        indicator information
    """
    # Add the base task if it's missing
    if "base" not in S.dtype.names:
        # Create a new np.recarray with an additional "base" data field
        S = rfn.append_fields(
            [S], names=[("base")], data=[np.ones(S.shape)], asrecarray=True
        )

    slice_names = S.dtype.names

    Y_dict: Dict[str, torch.Tensor] = dataloader.dataset.Y_dict  # type: ignore
    labels = Y_dict[base_task.name]

    for slice_name in slice_names:
        # Gather ind labels
        ind_labels = torch.LongTensor(S[slice_name])  # type: ignore

        # Mask out "inactive" pred_labels as specified by ind_labels
        pred_labels = labels.clone()
        pred_labels[~ind_labels.bool()] = -1

        ind_task_name = f"{base_task.name}_slice:{slice_name}_ind"
        pred_task_name = f"{base_task.name}_slice:{slice_name}_pred"

        # Update dataloaders
        Y_dict[ind_task_name] = ind_labels
        Y_dict[pred_task_name] = pred_labels


def convert_to_slice_tasks(base_task: Task, slice_names: List[str]) -> List[Task]:
    """Add slice labels to dataloader and creates new slice tasks (including base slice).

    Each slice will get two slice-specific heads:
    - an indicator head that learns to identify when DataPoints are in that slice
    - a predictor head that is trained on only members of that slice

    The base task's head is replaced by a master head that makes predictions based on
    a combination of the predictor heads' predictions that are weighted by the
    indicator heads' prediction confidences.

    WARNING: The current implementation pollutes the module_pool---the indicator task's
    module_pool includes predictor modules and vice versa since both are modified in
    place. This does not affect the result because the op sequences dictate which modules
    get used, and those do not include the extra modules. An alternative would be to
    make separate copies of the module pool for each, but that wastes time and memory
    extra copies of (potentially very large) modules that will be merged in a moment
    away in the model since they have the same name. We leave resolution of this issue
    for a future release.


    Parameters
    ----------
    base_task
        Task for which we are adding slice tasks. As noted in the WARNING, this task's
        module_pool will currently be modified in place for efficiency purposes.
    slice_names
        List of slice names corresponding to the columns of the slice matrix.

    Returns
    -------
    List[Task]
        Containins original base_task, pred/ind tasks for the base slice, and pred/ind
        tasks for each of the specified slice_names
    """

    if "base" not in slice_names:
        slice_names = slice_names + ["base"]

    slice_tasks: List[Task] = []

    # Keep track of all operations related to slice tasks
    slice_task_ops: List[Operation] = []

    # NOTE: We assume here that the last operation uses the head module
    # Identify base task head module
    head_module_op = base_task.op_sequence[-1]
    head_module = base_task.module_pool[head_module_op.module_name]

    if isinstance(head_module, nn.DataParallel):
        head_module = head_module.module

    neck_size = head_module.in_features
    assert isinstance(neck_size, int)
    base_task_cardinality = head_module.out_features
    assert isinstance(base_task_cardinality, int)

    # Remove the slice-unaware head module from module pool and op sequence
    del base_task.module_pool[head_module_op.module_name]
    body_flow = base_task.op_sequence[:-1]

    # Create slice indicator tasks
    for slice_name in slice_names:

        ind_task_name = f"{base_task.name}_slice:{slice_name}_ind"
        ind_head_module_name = f"{ind_task_name}_head"
        # Indicator head always predicts "in the slice or not", so is always binary
        ind_head_module = nn.Linear(neck_size, 2)

        # Create module_pool
        ind_module_pool = base_task.module_pool
        ind_module_pool[ind_head_module_name] = ind_head_module

        # Define operations for task head
        ind_head_op = Operation(
            module_name=ind_head_module_name, inputs=head_module_op.inputs
        )
        ind_task_ops = [ind_head_op]
        slice_task_ops.extend(ind_task_ops)

        # Create op sequence
        ind_op_sequence = list(body_flow) + list(ind_task_ops)

        # Create ind task
        ind_task = Task(
            name=ind_task_name,
            module_pool=ind_module_pool,
            op_sequence=ind_op_sequence,
            # NOTE: F1 by default because indicator task is often class imbalanced
            scorer=Scorer(metrics=["f1"]),
        )
        slice_tasks.append(ind_task)

    # Create slice predictor tasks
    shared_pred_head_module = nn.Linear(neck_size, base_task_cardinality)
    for slice_name in slice_names:

        pred_task_name = f"{base_task.name}_slice:{slice_name}_pred"

        pred_head_module_name = f"{pred_task_name}_head"
        pred_transform_module_name = f"{pred_task_name}_transform"
        pred_transform_module = nn.Linear(neck_size, neck_size)

        # Create module_pool
        # NOTE: See note in doc string about module_pool polution
        pred_module_pool = base_task.module_pool
        pred_module_pool[pred_transform_module_name] = pred_transform_module
        pred_module_pool[pred_head_module_name] = shared_pred_head_module

        # Define operations for task head
        pred_transform_op = Operation(
            module_name=pred_transform_module_name, inputs=head_module_op.inputs
        )
        pred_head_op = Operation(
            module_name=pred_head_module_name, inputs=[pred_transform_op.name]
        )
        pred_task_ops = [pred_transform_op, pred_head_op]
        slice_task_ops.extend(pred_task_ops)

        # Create op sequence
        pred_op_sequence = list(body_flow) + list(pred_task_ops)

        # Create pred task
        pred_task = Task(
            name=pred_task_name,
            module_pool=pred_module_pool,
            op_sequence=pred_op_sequence,
            scorer=base_task.scorer,
        )
        slice_tasks.append(pred_task)

    # Create master task
    master_task_name = base_task.name
    master_combiner_module_name = f"{base_task.name}_master_combiner"
    master_combiner_module = SliceCombinerModule()
    master_head_module_name = f"{base_task.name}_master_head"
    master_head_module = head_module

    # Create module_pool
    master_module_pool = nn.ModuleDict(
        {
            master_combiner_module_name: master_combiner_module,
            master_head_module_name: master_head_module,
        }
    )

    master_combiner_op = Operation(module_name=master_combiner_module_name, inputs=[])
    master_head_op = Operation(
        module_name=master_head_module_name, inputs=[master_combiner_op.name]
    )

    # NOTE: See note in doc string about module_pool polution

    # Create op_sequence
    master_op_sequence = (
        list(body_flow) + list(slice_task_ops) + [master_combiner_op, master_head_op]
    )

    master_task = Task(
        name=master_task_name,
        module_pool=master_module_pool,
        op_sequence=master_op_sequence,
        scorer=base_task.scorer,
    )
    return slice_tasks + [master_task]



"""Programmatic data set labeling: LF creation, models, and analysis utilities."""

from .analysis import LFAnalysis  # noqa: F401
from .apply.core import LFApplier  # noqa: F401
from .apply.pandas import PandasLFApplier  # noqa: F401
from .lf.core import LabelingFunction, labeling_function  # noqa: F401
from .utils import filter_unlabeled_dataframe  # noqa: F401



from typing import Any, Dict, List, Optional, Set, Union

import numpy as np
import pandas as pd
import torch
import torch.nn as nn

from ..analysis import Scorer
from ..classification import DictDataLoader, DictDataset, Operation, Task
from ..classification.data import DEFAULT_INPUT_DATA_KEY, DEFAULT_TASK_NAME
from ..classification.multitask_classifier import MultitaskClassifier

from .utils import add_slice_labels, convert_to_slice_tasks


class SliceAwareClassifier(MultitaskClassifier):
    """A slice-aware classifier that supports training + scoring on slice labels.

    NOTE: This model currently only supports binary classification.

    Parameters
    ----------
    base_architecture
        A network architecture that accepts input data and outputs a representation
    head_dim
        Output feature dimension of the base_architecture, and input dimension of the
        internal prediction head: ``nn.Linear(head_dim, 2)``.
    slice_names
        A list of slice names that the model will accept initialize as tasks
        and accept as corresponding labels
    scorer
        A Scorer to be used for initialization of the ``MultitaskClassifier`` superclass.
    **multitask_kwargs
        Arbitrary keyword arguments to be passed to the ``MultitaskClassifier`` superclass.

    Attributes
    ----------
    base_task
        A base ``snorkel.classification.Task`` that the model will learn.
        This becomes a ``master_head_module`` that combines slice tasks information.
        For more, see ``snorkel.slicing.convert_to_slice_tasks``.
    slice_names
        See above
    """

    def __init__(
        self,
        base_architecture: nn.Module,
        head_dim: int,
        slice_names: List[str],
        input_data_key: str = DEFAULT_INPUT_DATA_KEY,
        task_name: str = DEFAULT_TASK_NAME,
        scorer: Scorer = Scorer(metrics=["accuracy", "f1"]),
        **multitask_kwargs: Any,
    ) -> None:

        # Initialize module_pool with 1) base_architecture and 2) prediction_head
        # Assuming `head_dim` can be used to map base_architecture to prediction_head
        module_pool = nn.ModuleDict(
            {
                "base_architecture": base_architecture,
                "prediction_head": nn.Linear(head_dim, 2),
            }
        )

        # Create op_sequence from base_architecture -> prediction_head
        op_sequence = [
            Operation(
                name="input_op",
                module_name="base_architecture",
                inputs=[("_input_", input_data_key)],
            ),
            Operation(
                name="head_op", module_name="prediction_head", inputs=["input_op"]
            ),
        ]

        # Initialize base_task using specified base_architecture
        self.base_task = Task(
            name=task_name,
            module_pool=module_pool,
            op_sequence=op_sequence,
            scorer=scorer,
        )

        # Convert base_task to associated slice_tasks
        slice_tasks = convert_to_slice_tasks(self.base_task, slice_names)

        # Initialize a MultitaskClassifier with all slice_tasks
        model_name = f"{task_name}_sliceaware_classifier"
        super().__init__(tasks=slice_tasks, name=model_name, **multitask_kwargs)
        self.slice_names = slice_names

    def make_slice_dataloader(
        self, dataset: DictDataset, S: np.recarray, **dataloader_kwargs: Any
    ) -> DictDataLoader:
        """Create DictDataLoader with slice labels, initialized from specified dataset.

        Parameters
        ----------
        dataset
            A DictDataset that will be converted into a slice-aware dataloader
        S
            A [num_examples, num_slices] slice matrix indicating whether
            each example is in every slice
        slice_names
            A list of slice names corresponding to columns of ``S``

        dataloader_kwargs
            Arbitrary kwargs to be passed to DictDataLoader
            See ``DictDataLoader.__init__``.
        """

        # Base task must have corresponding labels in dataset
        if self.base_task.name not in dataset.Y_dict:  # type: ignore
            raise ValueError(
                f"Base task ({self.base_task.name}) labels missing from {dataset}"
            )

        # Initialize dataloader
        dataloader = DictDataLoader(dataset, **dataloader_kwargs)

        # Make dataloader slice-aware
        add_slice_labels(dataloader, self.base_task, S)

        return dataloader

    @torch.no_grad()
    def score_slices(
        self, dataloaders: List[DictDataLoader], as_dataframe: bool = False
    ) -> Union[Dict[str, float], pd.DataFrame]:
        """Scores appropriate slice labels using the overall prediction head.

        In other words, uses ``base_task`` (NOT ``slice_tasks``) to evaluate slices.

        In practice, we'd like to use a final prediction from a _single_ task head.
        To do so, ``self.base_task`` leverages reweighted slice representation to
        make a prediction. In this method, we remap all slice-specific ``pred``
        labels to ``self.base_task`` for evaluation.

        Parameters
        ----------
        dataloaders
            A list of DictDataLoaders to calculate scores for
        as_dataframe
            A boolean indicating whether to return results as pandas
            DataFrame (True) or dict (False)
        eval_slices_on_base_task
            A boolean indicating whether to remap slice labels to base task.
            Otherwise, keeps evaluation of slice labels on slice-specific heads.

        Returns
        -------
        Dict[str, float]
            A dictionary mapping metric¡ names to corresponding scores
            Metric names will be of the form "task/dataset/split/metric"
        """

        eval_mapping: Dict[str, Optional[str]] = {}
        # Collect all labels
        all_labels: Union[List, Set] = []
        for dl in dataloaders:
            all_labels.extend(dl.dataset.Y_dict.keys())  # type: ignore
        all_labels = set(all_labels)

        # By convention, evaluate on "pred" labels, not "ind" labels
        # See ``snorkel.slicing.utils.add_slice_labels`` for more about label creation
        for label in all_labels:
            if "pred" in label:
                eval_mapping[label] = self.base_task.name
            elif "ind" in label:
                eval_mapping[label] = None

        # Call score on the original remapped set of labels
        return super().score(
            dataloaders=dataloaders,
            remap_labels=eval_mapping,
            as_dataframe=as_dataframe,
        )



import os
import re
import pysbd
from pathlib import Path
import pandas as pd
from ekorpkit import eKonf
from glob import glob


class Pathobook:
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        self.input_path = self.args.input_path
        self.output_dir = Path(self.args.output_dir)
        self.chapter_dir = self.output_dir / "chapters"
        self.chapter_dir.mkdir(parents=True, exist_ok=True)
        self.norm_dir = self.output_dir / "normalized"
        self.norm_dir.mkdir(parents=True, exist_ok=True)
        self.output_file = self.output_dir / self.args.output_file
        # self.download()
        self.chapter_info = pd.read_csv(self.args.chapter_info_path, index_col=None)
        # self.normalizer = instantiate(self.args.normalize)

        if os.path.exists(self.output_file) and not self.args.force_download:
            print(f"output file {self.output_file} already exists, skipping")
        else:
            self.build()

    def build(self):
        self.split_chapters()
        self.make_clean_corpus()

    def split_chapters(self):
        contents = ""
        filepaths = glob(self.input_path, recursive=True)
        filepaths = [fp for fp in filepaths if Path(fp).is_file()]
        filepaths = sorted(filepaths)
        print(filepaths)
        self.norm_files = []
        for fp in filepaths:
            with open(fp, "r") as f:
                contents += f.read()

        print(contents[:100])
        self.chapters = {}
        self.chapter_files = {}
        for i, row in self.chapter_info.iterrows():
            str_ch, str_beg, str_end = row[["Chapter", "Beginning", "Ending"]]
            # str_beg = self.normalizer.normalize(str_beg)
            # str_end = self.normalizer.normalize(str_end)
            pos_beg = contents.find(str_beg)
            pos_end = contents.find(str_end)
            print(i + 1, pos_beg, pos_end)
            if pos_beg == -1:
                print(str_ch, str_beg)
            if pos_end == -1:
                print(str_ch, str_end)
            if pos_beg > -1:
                if pos_end > -1:
                    ch_txt = contents[pos_beg : pos_end + len(str_end) + 1]
                else:
                    ch_txt = contents[pos_beg:]
                filename = "chapter_{:0>2d}.txt".format(str_ch)
                self.chapters[str_ch] = ch_txt
                self.chapter_files[str_ch] = filename
                print(f"{filename}: {len(ch_txt)}")

            if pos_beg == -1 and pos_end == -1:
                print("failed to find chapter {}".format(str_ch))
                print(str_ch, str_beg, str_end)

    def make_clean_corpus(self):
        seg = pysbd.Segmenter(language="en", clean=False)
        pattern_remove_starting_with = "^\. ?"
        # pattern_split_para = '^\s*[\w+]+[\w+\s]*[^.:]$'
        pattern_split_para = "\n{2,}"
        # pattern_replace_hyphen = '(?<=\w)- '
        documents = []
        for chapter, doc in self.chapters.items():
            filename = self.chapter_files[chapter]
            print("processing {}".format(filename))

            # doc = re.sub(pattern_replace_hyphen, '-', doc)
            paras = re.split(pattern_split_para, doc, maxsplit=0, flags=re.MULTILINE)

            o_file = self.chapter_dir / filename
            out_paras = []
            for para in paras:
                if len(para) > 0:
                    sents = []
                    for sent in para.split("\n"):
                        sent = sent.strip()
                        sent = re.sub(pattern_remove_starting_with, "", sent)
                        # if len(sent) > 0:
                        #     seg_sent = seg.segment(sent)
                        #     for ss in seg_sent:
                        #         ss = re.sub(pattern_remove_starting_with, '', ss)
                        sents.append(sent)
                    out_paras.append("\n".join(sents))

            with o_file.open("w", encoding="utf-8") as fo:
                fo.write("\n\n".join(out_paras))
            doc = {"chapter": chapter, "text": "\n\n".join(out_paras)}
            documents.append(doc)
        df = pd.DataFrame(documents)
        df.to_csv(self.output_file, index=False)



import os
import codecs
import subprocess
from ekorpkit import eKonf


def rsync(**cfg):
    args = eKonf.to_config(cfg)
    os.makedirs(args.output_dir, exist_ok=True)

    options = codecs.decode(args.options, "unicode_escape")
    cmd = ["rsync", options, args.source, args.output_dir]
    if not os.listdir(args.output_dir) or args.force_download:
        print(f" >> rsyncing {args.source} to {args.output_dir}")
        subprocess.run(" ".join(cmd), shell=True)



import logging
import os
import json
import pandas as pd
from ekorpkit import eKonf
from ekorpkit.io.fetch.base import BaseFetcher


log = logging.getLogger(__name__)


class ECOS(BaseFetcher):
    def __init__(self, **args):
        self.args = eKonf.to_config(args)
        super().__init__(**args)

        self._ecos_api_key = self.args.get("ecos_api_key")
        self._api = self.args.get("api")

        self.series_id = self.args.get("series_id")
        self.value_column = self.args.get("value_column") or "value"
        self.start_date = self.args.get("start_date")
        self.end_date = self.args.get("end_date")
        self.cycle = self.args.get("cycle")
        self.eval_columns = self.args.get("pipeline").get("eval_columns")

        self.output_dir = self.args["output_dir"]
        os.makedirs(self.output_dir, exist_ok=True)
        self.output_file = self.args["output_file"]

        if self.auto.load:
            self.load(
                series_id=self.series_id,
                start_date=self.start_date,
                end_date=self.end_date,
                cycle=self.cycle,
            )

    def get(self, series_id, start_date=None, end_date=None, cycle="", **kwargs):
        codes = [None, None, None, None, None]
        if "/" in series_id:
            series_id = series_id.split("/")
            codes[: len(series_id)] = series_id
        else:
            codes[0] = series_id
        data, _ = self.get_ecos(
            service="StatisticSearch",
            stat_code=codes[0],
            cycle=cycle,
            start_date=start_date,
            end_date=end_date,
            item_code1=codes[1],
            item_code2=codes[2],
            item_code3=codes[3],
            item_code4=codes[4],
            **kwargs,
        )
        return data

    def get_ecos(
        self,
        service,
        stat_code,
        cycle="",
        start_date="",
        end_date="",
        start_num=None,
        end_num=None,
        item_code1=None,
        item_code2=None,
        item_code3=None,
        item_code4=None,
        lang=None,
        format=None,
        **kwargs,
    ):
        import requests

        lang = lang or self._api.lang
        format = format or self._api.format
        start_num = start_num or self._api.start_num
        end_num = end_num or self._api.end_num

        url = "{}/{}/{}/{}/{}/{}/{}/{}/{}/{}/{}".format(
            self._api.base_url,
            service,
            self._ecos_api_key,
            format,
            lang,
            start_num,
            end_num,
            stat_code,
            cycle,
            start_date,
            end_date,
        )
        if item_code1 is not None:
            item_code2 = item_code2 or "?"
            item_code3 = item_code3 or "?"
            item_code4 = item_code4 or "?"

            url += "/{}/{}/{}/{}".format(item_code1, item_code2, item_code3, item_code4)

        response = requests.get(url).content.decode("utf-8")
        response = json.loads(response)
        if service in response:
            data = pd.DataFrame(response[service]["row"])
            num_rows = response[service]["list_total_count"]

            return data, num_rows
        else:
            log.info(f"No data found for {service}")
            if self.verbose:
                print(response)
            return None, None

    def load(
        self,
        series_id,
        start_date=None,
        end_date=None,
        cycle=None,
        filename=None,
        expressions=None,
        reset_index=False,
    ):
        if isinstance(series_id, str):
            series_id = [series_id]
        elif not isinstance(series_id, list):
            series_id = []

        if filename is None:
            file_name = "_".join(series_id).replace("/", "_")
            filename = f"{file_name}.parquet"

        if start_date is None:
            start_date = self.start_date
        if end_date is None:
            end_date = self.end_date
        if cycle is None:
            cycle = self.cycle

        if isinstance(self.eval_columns, dict):
            self.eval_columns["expressions"] = expressions

        if self.verbose:
            print(f"Loading {series_id} from {start_date} to {end_date}")

        filepath = os.path.join(self.output_dir, filename)
        if not os.path.exists(filepath) or self.force.download:
            self._data = self._load_series(
                series_id, start_date, end_date, cycle, reset_index
            )
            eKonf.save_data(self.data, filepath, verbose=self.verbose)
        else:
            log.info(f"{filepath} already exists.")
            self._data = eKonf.load_data(filepath, verbose=self.verbose)
        return self.data.copy()

    def _load_series(
        self,
        series_ids=None,
        start_date=None,
        end_date=None,
        cycle=None,
        reset_index=False,
        **kwargs,
    ):

        _dfs = []
        for series_id in series_ids:
            df = self.get(
                series_id=series_id,
                start_date=start_date,
                end_date=end_date,
                cycle=cycle,
                **kwargs,
            )
            df = eKonf.pipe(df, self.eval_columns)
            if len(series_ids) > 1:
                df["series_id"] = series_id
            if self.verbose:
                print(df.head())
            _dfs.append(df)

        df = pd.concat(_dfs)
        if isinstance(reset_index, str) and reset_index in df.columns:
            df = df.set_index(reset_index)
        elif reset_index:
            df = df.reset_index()
        return df



import itertools
import json
import math
import os
import pandas as pd
import re
import requests
import tempfile
import zipfile
import logging
from bs4 import BeautifulSoup
from datetime import datetime
from requests.adapters import HTTPAdapter
from requests.exceptions import (
    RequestException,
    HTTPError,
    ConnectionError,
    Timeout,
    RetryError,
)
from tqdm.auto import tqdm
from typing import List
from urllib3.util import Retry
from ekorpkit import eKonf

try:
    from html.parser import NewTypeOfFancyException as HTMLParseError
except ImportError:  # Python 3.5+

    class HTMLParseError(Exception):
        pass


logging.getLogger("urllib3").setLevel(logging.CRITICAL)
log = logging.getLogger(__name__)


def download_filings(**args):
    """
    The main method iterates all over the tsv index files that are generated
    and calls a scrap method for each one of them.
    """
    config = eKonf.to_config(args)

    output_dir = config["output_dir"]
    raw_filings_dir = os.path.join(output_dir, config["raw_filings_dir"])
    indices_dir = os.path.join(output_dir, config["indices_dir"])
    filings_metadata_filepath = os.path.join(
        output_dir, config["filings_metadata_file"]
    )
    companies_info_filepath = os.path.join(output_dir, config["companies_info_file"])
    if len(config["filing_types"]) == 0:
        log.info(f"Please provide at least one filing type")
        exit()

    # If the indices and/or download folder doesn't exist, create them
    if not os.path.isdir(indices_dir):
        os.mkdir(indices_dir)
    if not os.path.isdir(raw_filings_dir):
        os.mkdir(raw_filings_dir)

    if not os.path.isfile(companies_info_filepath):
        with open(companies_info_filepath, "w") as f:
            json.dump(obj={}, fp=f)

    download_indices(
        start_year=config["start_year"],
        end_year=config["end_year"],
        quarters=config["quarters"],
        skip_present_indices=config["skip_present_indices"],
        indices_folder=indices_dir,
        user_agent=config["user_agent"],
        base_url=config["base_url"],
    )

    # Filter out years that are not related
    tsv_filenames = []
    for year in range(config["start_year"], config["end_year"] + 1):
        for quarter in config["quarters"]:
            filepath = os.path.join(indices_dir, f"{year}_QTR{quarter}.tsv")

            if os.path.isfile(filepath):
                tsv_filenames.append(filepath)

    # Get the indices that are specific to your needs
    df = get_specific_indices(
        tsv_filenames=tsv_filenames,
        filing_types=config["filing_types"],
        cik_tickers=config["cik_tickers"],
        user_agent=config["user_agent"],
    )

    old_df = None
    if os.path.exists(filings_metadata_filepath):
        old_df = []
        series_to_download = []
        log.info(f"\nReading filings metadata...\n")

        for _, series in pd.read_csv(filings_metadata_filepath, dtype=str).iterrows():
            if os.path.exists(os.path.join(raw_filings_dir, series["filename"])):
                old_df.append((series.to_frame()).T)
        old_df = pd.concat(old_df) if (len(old_df) > 1) else old_df[0]

        for _, series in tqdm(df.iterrows(), total=len(df), ncols=100):
            if len(old_df[old_df["html_index"] == series["html_index"]]) == 0:
                series_to_download.append((series.to_frame()).T)

        if len(series_to_download) == 0:
            log.info(
                f"\nThere are no more filings to download for the given years, quarters and companies"
            )
            exit()

        df = (
            pd.concat(series_to_download)
            if (len(series_to_download) > 1)
            else series_to_download[0]
        )

    # Make a list for each series of them
    list_of_series = []
    for i in range(len(df)):
        list_of_series.append(df.iloc[i])

    log.info(f"\nDownloading {len(df)} filings...\n")

    final_series = []
    for series in tqdm(list_of_series, ncols=100):
        series = scrap(
            companies_info_filepath=companies_info_filepath,
            series=series,
            filing_types=config["filing_types"],
            raw_filings_folder=raw_filings_dir,
            user_agent=config["user_agent"],
        )
        if series is not None:
            final_series.append((series.to_frame()).T)
            final_df = (
                pd.concat(final_series) if (len(final_series) > 1) else final_series[0]
            )
            final_df = pd.concat([old_df, final_df])
            final_df.to_csv(filings_metadata_filepath, index=False, header=True)

    log.info(f"\nFinal dataframe exported to {filings_metadata_filepath}")


def download_indices(
    start_year: int,
    end_year: int,
    quarters: List,
    skip_present_indices: bool,
    indices_folder: str,
    user_agent: str,
    base_url: str = "https://www.sec.gov/Archives/edgar/full-index/",
):
    log.info("Downloading EDGAR Index files")

    for quarter in quarters:
        if quarter not in [1, 2, 3, 4]:
            raise Exception(f'Invalid quarter "{quarter}"')

    first_iteration = True
    while True:
        failed_indices = []
        for year in range(start_year, end_year + 1):
            for quarter in quarters:
                if year == datetime.now().year and quarter > math.ceil(
                    datetime.now().month / 3
                ):
                    break
                index_filename = f"{year}_QTR{quarter}.tsv"
                if skip_present_indices and os.path.exists(
                    os.path.join(indices_folder, index_filename)
                ):
                    if first_iteration:
                        log.info(f"Skipping {index_filename}")
                    continue

                url = f"{base_url}/{year}/QTR{quarter}/master.zip"

                with tempfile.TemporaryFile(mode="w+b") as tmp:
                    session = requests.Session()
                    try:
                        request = requests_retry_session(
                            retries=5, backoff_factor=0.2, session=session
                        ).get(url=url, headers={"User-agent": user_agent})
                    except requests.exceptions.RetryError as e:
                        log.info(f'Failed downloading "{index_filename}" - {e}')
                        failed_indices.append(index_filename)
                        continue

                    tmp.write(request.content)
                    with zipfile.ZipFile(tmp).open("master.idx") as f:
                        lines = [
                            line.decode("latin-1")
                            for line in itertools.islice(f, 11, None)
                        ]
                        lines = [
                            line.strip()
                            + "|"
                            + line.split("|")[-1].replace(".txt", "-index.html")
                            for line in lines
                        ]

                    with open(
                        os.path.join(indices_folder, index_filename),
                        "w+",
                        encoding="utf-8",
                    ) as f:
                        f.write("".join(lines))
                        log.info(f"{index_filename} downloaded")

        first_iteration = False
        if len(failed_indices) > 0:
            log.info(f"Could not download the following indices:\n{failed_indices}")
            user_input = input("Retry (Y/N): ")
            if user_input in ["Y", "y", "yes"]:
                log.info(f"Retry downloading failed indices")
            else:
                break
        else:
            break


def get_specific_indices(
    tsv_filenames,
    filing_types,
    user_agent,
    cik_tickers=None,
):
    """
    Loops through all the indexes and keeps only the rows/Series for the specific filing types
    :param tsv_filenames: the indices filenames
    :param filing_types: list of filing types to download. e.g. ['10-K', '10-K405', '10-KT']
    :param user_agent: the User-agent that will be declared to SEC EDGAR
    :param cik_tickers: list of CIKs or Tickers
    :return: a final dataframe which has Series only for the specific indices
    """

    ciks = []

    if cik_tickers is not None:
        if isinstance(cik_tickers, str):
            if os.path.exists(cik_tickers) and os.path.isfile(
                cik_tickers
            ):  # If filepath
                with open(cik_tickers) as f:
                    cik_tickers = [
                        line.strip() for line in f.readlines() if line.strip() != ""
                    ]
            else:
                log.debug(f"Please provide a valid cik_ticker file path")
                exit()

    if isinstance(cik_tickers, List) and len(cik_tickers):
        company_tickers_url = "https://www.sec.gov/files/company_tickers.json"

        session = requests.Session()
        try:
            request = requests_retry_session(
                retries=5, backoff_factor=0.2, session=session
            ).get(url=company_tickers_url, headers={"User-agent": user_agent})
        except (
            RequestException,
            HTTPError,
            ConnectionError,
            Timeout,
            RetryError,
        ) as err:
            log.info(f'Failed downloading "{company_tickers_url}" - {err}')
            exit()

        company_tickers = json.loads(request.content)
        ticker2cik = {
            company["ticker"]: company["cik_str"]
            for company in company_tickers.values()
        }
        ticker2cik = dict(sorted(ticker2cik.items(), key=lambda item: item[0]))

        for c_t in cik_tickers:
            if isinstance(c_t, int) or c_t.isdigit():  # If CIK
                ciks.append(str(c_t))
            else:
                if c_t in ticker2cik:
                    ciks.append(str(ticker2cik[c_t]))  # If Ticker
                else:
                    log.debug(f'Could not find CIK for "{c_t}"')

    dfs_list = []

    for filepath in tsv_filenames:

        # Load the index file
        df = pd.read_csv(
            filepath,
            sep="|",
            header=None,
            dtype=str,
            names=[
                "CIK",
                "Company",
                "Type",
                "Date",
                "complete_text_file_link",
                "html_index",
                "Filing Date",
                "Period of Report",
                "SIC",
                "htm_file_link",
                "State of Inc",
                "State location",
                "Fiscal Year End",
                "filename",
            ],
        )

        df["complete_text_file_link"] = "https://www.sec.gov/Archives/" + df[
            "complete_text_file_link"
        ].astype(str)
        df["html_index"] = "https://www.sec.gov/Archives/" + df["html_index"].astype(
            str
        )

        # Filter by filing type
        df = df[df.Type.isin(filing_types)]

        # Filter by CIK
        if len(ciks):
            df = df[(df.CIK.isin(ciks))]

        dfs_list.append(df)

    return pd.concat(dfs_list) if (len(dfs_list) > 1) else dfs_list[0]


def scrap(
    companies_info_filepath, filing_types, series, raw_filings_folder, user_agent
):
    """
    Scraps the EDGAR HTML indexes
    :param companies_info_filepath: path to companies info file
    :param filing_types: list of filing types to download
    :param series: A single series with info for specific filings
    :param raw_filings_folder: Raw filings folder path
    :param user_agent: the User-agent that will be declared to SEC EDGAR
    :return: the .htm or .txt files
    """

    html_index = series["html_index"]

    # Create a BeautifulSoup instance using the 'lxml' parser
    try:
        retries_exceeded = True
        for _ in range(5):
            session = requests.Session()
            request = requests_retry_session(
                retries=5, backoff_factor=0.2, session=session
            ).get(url=html_index, headers={"User-agent": user_agent})

            if (
                "will be managed until action is taken to declare your traffic."
                not in request.text
            ):
                retries_exceeded = False
                break

        if retries_exceeded:
            log.debug(f'Retries exceeded, could not download "{html_index}"')
            return None

    except (RequestException, HTTPError, ConnectionError, Timeout, RetryError) as err:
        log.debug(
            f"Request for {html_index} failed due to network-related error: {err}"
        )
        return None

    soup = BeautifulSoup(request.content, "lxml")

    # Crawl the soup and search it later for the Period of Report
    try:
        list_of_forms = soup.find_all("div", {"class": ["infoHead", "info"]})
    except (HTMLParseError, Exception) as e:
        list_of_forms = None

    period_of_report = None
    for form in list_of_forms:
        if form.attrs["class"][0] == "infoHead" and form.text == "Filing Date":
            series["Filing Date"] = form.nextSibling.nextSibling.text

        if form.attrs["class"][0] == "infoHead" and form.text == "Period of Report":
            period_of_report = form.nextSibling.nextSibling.text
            series["Period of Report"] = period_of_report

    if period_of_report is None:
        log.debug(f'Can not crawl "Period of Report" for {html_index}')
        return None

    # Assign metadata to dataframe
    try:
        company_info = (
            soup.find("div", {"class": ["companyInfo"]})
            .find("p", {"class": ["identInfo"]})
            .text
        )
    except (HTMLParseError, Exception) as e:
        company_info = None

    try:
        for info in company_info.split("|"):
            info_splits = info.split(":")
            if info_splits[0].strip() in [
                "State of Incorp.",
                "State of Inc.",
                "State of Incorporation.",
            ]:
                series["State of Inc"] = info_splits[1].strip()
            if info_splits[0].strip() == ["State location"]:
                series["State location"] = info_splits[1].strip()
    except (ValueError, Exception) as e:
        pass

    fiscal_year_end_regex = re.search(r"Fiscal Year End: *(\d{4})", company_info)
    if fiscal_year_end_regex is not None:
        series["Fiscal Year End"] = fiscal_year_end_regex.group(1)

    # Crawl for the Sector Industry Code (SIC)
    try:
        sic = soup.select_one('.identInfo a[href*="SIC"]')
        if sic is not None:
            series["SIC"] = sic.text
    except (HTMLParseError, Exception) as e:
        pass

    # https://www.sec.gov/cgi-bin/browse-edgar?CIK=0001000228
    # https://data.sec.gov/submissions/CIK0001000228.json
    with open(companies_info_filepath) as f:
        company_info_dict = json.load(fp=f)

    cik = series["CIK"]
    if cik not in company_info_dict:
        company_url = f"https://www.sec.gov/cgi-bin/browse-edgar?CIK={cik}"
        try:
            retries_exceeded = True
            for _ in range(5):
                session = requests.Session()
                request = requests_retry_session(
                    retries=5, backoff_factor=0.2, session=session
                ).get(url=company_url, headers={"User-agent": user_agent})

                if (
                    "will be managed until action is taken to declare your traffic."
                    not in request.text
                ):
                    retries_exceeded = False
                    break

            if retries_exceeded:
                log.debug(f'Retries exceeded, could not download "{company_url}"')
                return None

        except (
            RequestException,
            HTTPError,
            ConnectionError,
            Timeout,
            RetryError,
        ) as err:
            log.debug(
                f"Request for {company_url} failed due to network-related error: {err}"
            )
            return None

        company_info_dict[cik] = {
            "Company Name": None,
            "SIC": None,
            "State location": None,
            "State of Inc": None,
            "Fiscal Year End": None,
        }
        company_info_soup = BeautifulSoup(request.content, "lxml")

        company_info = company_info_soup.find("div", {"class": ["companyInfo"]})
        if company_info is not None:
            company_info_dict[cik]["Company Name"] = str(
                company_info.find("span", {"class": ["companyName"]}).contents[0]
            ).strip()
            company_info_contents = company_info.find(
                "p", {"class": ["identInfo"]}
            ).contents

            for idx, content in enumerate(company_info_contents):
                if ";SIC=" in str(content):
                    company_info_dict[cik]["SIC"] = content.text
                if ";State=" in str(content):
                    company_info_dict[cik]["State location"] = content.text
                if "State of Inc" in str(content):
                    company_info_dict[cik]["State of Inc"] = company_info_contents[
                        idx + 1
                    ].text
                if "Fiscal Year End" in str(content):
                    company_info_dict[cik]["Fiscal Year End"] = str(content).split()[-1]

        with open(companies_info_filepath, "w") as f:
            json.dump(obj=company_info_dict, fp=f, indent=4)

    if pd.isna(series["SIC"]):
        series["SIC"] = company_info_dict[cik]["SIC"]
    if pd.isna(series["State of Inc"]):
        series["State of Inc"] = company_info_dict[cik]["State of Inc"]
    if pd.isna(series["State location"]):
        series["State location"] = company_info_dict[cik]["State location"]
    if pd.isna(series["Fiscal Year End"]):
        series["Fiscal Year End"] = company_info_dict[cik]["Fiscal Year End"]

    # Crawl the soup for the financial files
    try:
        all_tables = soup.find_all("table")
    except (HTMLParseError, Exception) as e:
        return None

    """
	Tables are of 2 kinds. 
	The 'Document Format Files' table contains all the htms, jpgs, pngs and txts for the reports.
	The 'Data Format Files' table contains all the xml instances that contain structured information.
	"""
    for table in all_tables:

        # Get the htm/html/txt files
        if table.attrs["summary"] == "Document Format Files":
            htm_file_link, complete_text_file_link, link_to_download = None, None, None
            filing_type = None

            for tr in table.find_all("tr")[1:]:
                # If it's the specific document type (e.g. 10-K)
                if tr.contents[7].text in filing_types:
                    filing_type = tr.contents[7].text
                    if tr.contents[5].contents[0].attrs["href"].split(".")[-1] in [
                        "htm",
                        "html",
                    ]:
                        htm_file_link = (
                            "https://www.sec.gov"
                            + tr.contents[5].contents[0].attrs["href"]
                        )
                        series["htm_file_link"] = str(htm_file_link)
                        break

                # Else get the complete submission text file
                elif tr.contents[3].text == "Complete submission text file":
                    filing_type = series["Type"]
                    complete_text_file_link = (
                        "https://www.sec.gov" + tr.contents[5].contents[0].attrs["href"]
                    )
                    series["complete_text_file_link"] = str(complete_text_file_link)
                    break

            if htm_file_link is not None:
                # In case of iXBRL documents, a slight URL modification is required
                if "ix?doc=/" in htm_file_link:
                    link_to_download = htm_file_link.replace("ix?doc=/", "")
                    series["htm_file_link"] = link_to_download
                    file_extension = "htm"
                else:
                    link_to_download = htm_file_link
                    file_extension = htm_file_link.split(".")[-1]

            elif complete_text_file_link is not None:
                link_to_download = complete_text_file_link
                file_extension = link_to_download.split(".")[-1]

            if link_to_download is not None:
                filing_type = re.sub(r"[\-/\\]", "", filing_type)
                accession_num = (
                    series["complete_text_file_link"].split(os.sep)[-1].split(".")[0]
                )
                filename = f"{str(series['CIK'])}_{filing_type}_{period_of_report[:4]}_{accession_num}.{file_extension}"

                # Download the file
                success = download(
                    url=link_to_download,
                    filename=filename,
                    download_folder=raw_filings_folder,
                    user_agent=user_agent,
                )
                if success:
                    series["filename"] = filename
                else:
                    return None
            else:
                return None

    return series


def download(url, filename, download_folder, user_agent):
    """
    Downloads the filing to the specified directory with the naming convention below:
    <CIK-KEY_YEAR_FILING-TYPE.EXTENSION_TYPE> (e.g.: 1000229_2018_10K.html)
    :param url: The URL to download
    :param filename: The Central Index Key (CIK) of the company
    :param download_folder:
    :param user_agent: the User-agent that will be declared to SEC EDGAR

    Note that we save files based on the years that they report to
    Most companies submit their reports on the end of December of the current year (2021 for example)
    However, if a company submits its report on the start of the next year (2022), then
    this will be saved as COMPANY_CIK_FILING-TYPE_2022.htm
    """

    filepath = os.path.join(download_folder, filename)

    try:
        retries_exceeded = True
        for _ in range(5):
            session = requests.Session()
            request = requests_retry_session(
                retries=5, backoff_factor=0.2, session=session
            ).get(url=url, headers={"User-agent": user_agent})
            # request = requests.get(html_index, headers={'User-Agent': ua.random})

            if (
                "will be managed until action is taken to declare your traffic."
                not in request.text
            ):
                retries_exceeded = False
                break

        if retries_exceeded:
            log.debug(f'Retries exceeded, could not download "{filename}" - "{url}"')
            return False

    except (RequestException, HTTPError, ConnectionError, Timeout, RetryError) as err:
        log.debug(f"Request for {url} failed due to network-related error: {err}")
        return False

    with open(filepath, "wb") as f:
        f.write(request.content)

    # logger.info(f'Successfully downloaded filing: {filename} - {url}')

    # Check that MD5 hash is correct
    # if hashlib.md5(open(filepath, 'rb').read()).hexdigest() != headers._headers[1][1].strip('"'):
    # 	logger.info(f'Wrong MD5 hash for file: {abs_filename} - {url}')

    return True


def requests_retry_session(
    retries=5,
    backoff_factor=0.5,
    status_forcelist=(400, 401, 403, 500, 502, 503, 504, 505),
    session=None,
):
    """
    Retries the HTTP GET method in case of some specific HTTP errors.

    :param retries: Time of retries
    :param backoff_factor: The amount of delay after each retry
    :param status_forcelist: The error codes that the script should retry; Otherwise, it won't retry
    :param session: the requests session
    :return: the new session
    """
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



import hashlib
from typing import Dict, List, Optional

import numpy as np


def _hash(i: int) -> int:
    """Deterministic hash function."""
    byte_string = str(i).encode("utf-8")
    return int(hashlib.sha1(byte_string).hexdigest(), 16)


def probs_to_preds(
    probs: np.ndarray, tie_break_policy: str = "random", tol: float = 1e-5
) -> np.ndarray:
    """Convert an array of probabilistic labels into an array of predictions.

    Policies to break ties include:
    "abstain": return an abstain vote (-1)
    "true-random": randomly choose among the tied options
    "random": randomly choose among tied option using deterministic hash

    NOTE: if tie_break_policy="true-random", repeated runs may have slightly different results due to difference in broken ties

    Parameters
    ----------
    prob
        A [num_datapoints, num_classes] array of probabilistic labels such that each
        row sums to 1.
    tie_break_policy
        Policy to break ties when converting probabilistic labels to predictions
    tol
        The minimum difference among probabilities to be considered a tie

    Returns
    -------
    np.ndarray
        A [n] array of predictions (integers in [0, ..., num_classes - 1])

    Examples
    --------
    >>> probs_to_preds(np.array([[0.5, 0.5, 0.5]]), tie_break_policy="abstain")
    array([-1])
    >>> probs_to_preds(np.array([[0.8, 0.1, 0.1]]))
    array([0])
    """
    num_datapoints, num_classes = probs.shape
    if num_classes <= 1:
        raise ValueError(
            f"probs must have probabilities for at least 2 classes. "
            f"Instead, got {num_classes} classes."
        )

    Y_pred = np.empty(num_datapoints)
    diffs = np.abs(probs - probs.max(axis=1).reshape(-1, 1))

    for i in range(num_datapoints):
        max_idxs = np.where(diffs[i, :] < tol)[0]
        if len(max_idxs) == 1:
            Y_pred[i] = max_idxs[0]
        # Deal with "tie votes" according to the specified policy
        elif tie_break_policy == "random":
            Y_pred[i] = max_idxs[_hash(i) % len(max_idxs)]
        elif tie_break_policy == "true-random":
            Y_pred[i] = np.random.choice(max_idxs)
        elif tie_break_policy == "abstain":
            Y_pred[i] = -1
        else:
            raise ValueError(
                f"tie_break_policy={tie_break_policy} policy not recognized."
            )
    return Y_pred.astype(np.int_)


def preds_to_probs(preds: np.ndarray, num_classes: int) -> np.ndarray:
    """Convert an array of predictions into an array of probabilistic labels.

    Parameters
    ----------
    pred
        A [num_datapoints] or [num_datapoints, 1] array of predictions

    Returns
    -------
    np.ndarray
        A [num_datapoints, num_classes] array of probabilistic labels with probability
        of 1.0 in the column corresponding to the prediction
    """
    if np.any(preds < 0):
        raise ValueError("Could not convert abstained vote to probability")
    return np.eye(num_classes)[preds.squeeze()]


def to_int_label_array(X: np.ndarray, flatten_vector: bool = True) -> np.ndarray:
    """Convert an array to a (possibly flattened) array of ints.

    Cast all values to ints and possibly flatten [n, 1] arrays to [n].
    This method is typically used to sanitize labels before use with analysis tools or
    metrics that expect 1D arrays as inputs.

    Parameters
    ----------
    X
        An array to possibly flatten and possibly cast to int
    flatten_vector
        If True, flatten array into a 1D array

    Returns
    -------
    np.ndarray
        The converted array

    Raises
    ------
    ValueError
        Provided input could not be converted to an np.ndarray
    """
    if np.any(np.not_equal(np.mod(X, 1), 0)):
        raise ValueError("Input contains at least one non-integer value.")
    X = X.astype(np.dtype(int))
    # Correct shape
    if flatten_vector:
        X = X.squeeze()
        if X.ndim == 0:
            X = np.expand_dims(X, 0)
        if X.ndim != 1:
            raise ValueError("Input could not be converted to 1d np.array")
    return X


def filter_labels(
    label_dict: Dict[str, Optional[np.ndarray]], filter_dict: Dict[str, List[int]]
) -> Dict[str, np.ndarray]:
    """Filter out examples from arrays based on specified labels to filter.

    The most common use of this method is to remove examples whose gold label is
    unknown (marked with a -1) or examples whose predictions were abstains (also -1)
    before calculating metrics.

    NB: If an example matches the filter criteria for any label set, it will be removed
    from all label sets (so that the returned arrays are of the same size and still
    aligned).

    Parameters
    ----------
    label_dict
        A mapping from label set name to the array of labels
        The arrays in a label_dict.values() are assumed to be aligned
    filter_dict
        A mapping from label set name to the labels that should be filtered out for
        that label set

    Returns
    -------
    Dict[str, np.ndarray]
        A mapping with the same keys as label_dict but with filtered arrays as values

    Example
    -------
    >>> golds = np.array([-1, 0, 0, 1, 0])
    >>> preds = np.array([0, 0, 0, 1, -1])
    >>> filtered = filter_labels(
    ...     label_dict={"golds": golds, "preds": preds},
    ...     filter_dict={"golds": [-1], "preds": [-1]}
    ... )
    >>> filtered["golds"]
    array([0, 0, 1])
    >>> filtered["preds"]
    array([0, 0, 1])
    """
    masks = []
    for label_name, filter_values in filter_dict.items():
        label_array: Optional[np.ndarray] = label_dict.get(label_name)
        if label_array is not None:
            # _get_mask requires not-null input
            masks.append(_get_mask(label_array, filter_values))
    mask = (np.multiply(*masks) if len(masks) > 1 else masks[0]).squeeze()

    filtered = {}
    for label_name, label_array in label_dict.items():
        filtered[label_name] = label_array[mask] if label_array is not None else None
    return filtered


def _get_mask(label_array: np.ndarray, filter_values: List[int]) -> np.ndarray:
    """Return a boolean mask marking which labels are not in filter_values.

    Parameters
    ----------
    label_array
        An array of labels
    filter_values
        A list of values that should be filtered out of the label array

    Returns
    -------
    np.ndarray
        A boolean mask indicating whether to keep (1) or filter (0) each example
    """
    mask: np.ndarray = np.ones_like(label_array).astype(bool)
    for value in filter_values:
        mask *= np.where(label_array != value, 1, 0).astype(bool)
    return mask



from typing import Iterable, List, Tuple

import networkx as nx


def get_clique_tree(nodes: Iterable[int], edges: List[Tuple[int, int]]) -> nx.Graph:
    """
    Given a set of int nodes i and edges (i,j), returns a clique tree.

    Clique tree is an object G for which:
    - G.nodes[i]['members'] contains the set of original nodes in the ith
        maximal clique
    - G[i][j]['members'] contains the set of original nodes in the seperator
        set between maximal cliques i and j

    Note: This method is currently only implemented for chordal graphs; TODO:
    add a step to triangulate non-chordal graphs.

    Parameters
    ----------
    nodes
        A list of nodes indices
    edges
        A list of tuples, where each tuple has indices for connected nodes

    Returns
    -------
    networkx.Graph
        An object G representing clique tree
    """
    # Form the original graph G1
    G1 = nx.Graph()
    G1.add_nodes_from(nodes)
    G1.add_edges_from(edges)

    # Check if graph is chordal
    # TODO: Add step to triangulate graph if not
    if not nx.is_chordal(G1):
        raise NotImplementedError("Graph triangulation not implemented.")

    # Create maximal clique graph G2
    # Each node is a maximal clique C_i
    # Let w = |C_i \cap C_j|; C_i, C_j have an edge with weight w if w > 0
    G2 = nx.Graph()
    for i, c in enumerate(nx.chordal_graph_cliques(G1)):
        G2.add_node(i, members=c)
    for i in G2.nodes:
        for j in G2.nodes:
            S = G2.nodes[i]["members"].intersection(G2.nodes[j]["members"])
            w = len(S)
            if w > 0:
                G2.add_edge(i, j, weight=w, members=S)

    # Return a minimum spanning tree of G2
    return nx.minimum_spanning_tree(G2)



from collections import defaultdict
from typing import Any, Callable, Dict, List, Tuple

from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from .utils import list_to_tensor

XDict = Dict[str, Any]
YDict = Dict[str, Tensor]
Batch = Tuple[XDict, YDict]

# Default string names for initializing a DictDataset
DEFAULT_INPUT_DATA_KEY = "input_data"
DEFAULT_DATASET_NAME = "SnorkelDataset"
DEFAULT_TASK_NAME = "task"


class DictDataset(Dataset):
    """A dataset where both the data fields and labels are stored in as dictionaries.

    Parameters
    ----------
    name
        The name of the dataset (e.g., this will be used to report metrics on a
        per-dataset basis)
    split
        The name of the split that the data in this object represents
    X_dict
        A map from field name to values (e.g., {"tokens": ..., "uids": ...})
    Y_dict
        A map from task name to its corresponding set of labels

    Raises
    ------
    ValueError
        All values in the ``Y_dict`` must be of type torch.Tensor

    Attributes
    ----------
    name
        See above
    split
        See above
    X_dict
        See above
    Y_dict
        See above
    """

    def __init__(self, name: str, split: str, X_dict: XDict, Y_dict: YDict) -> None:
        self.name = name
        self.split = split
        self.X_dict = X_dict
        self.Y_dict = Y_dict

        for name, label in self.Y_dict.items():
            if not isinstance(label, Tensor):
                raise ValueError(
                    f"Label {name} should be torch.Tensor, not {type(label)}."
                )

    def __getitem__(self, index: int) -> Tuple[XDict, YDict]:
        x_dict = {name: feature[index] for name, feature in self.X_dict.items()}
        y_dict = {name: label[index] for name, label in self.Y_dict.items()}
        return x_dict, y_dict

    def __len__(self) -> int:
        try:
            return len(next(iter(self.Y_dict.values())))  # type: ignore
        except StopIteration:
            return 0

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"(name={self.name}, "
            f"X_keys={list(self.X_dict.keys())}, "
            f"Y_keys={list(self.Y_dict.keys())})"
        )

    @classmethod
    def from_tensors(
        cls,
        X_tensor: Tensor,
        Y_tensor: Tensor,
        split: str,
        input_data_key: str = DEFAULT_INPUT_DATA_KEY,
        task_name: str = DEFAULT_TASK_NAME,
        dataset_name: str = DEFAULT_DATASET_NAME,
    ) -> "DictDataset":
        """Initialize a ``DictDataset`` from PyTorch Tensors.

        Parameters
        ----------
        X_tensor
            Input data of shape [num_examples, feature_dim]
        Y_tensor
            Labels of shape [num_samples, num_classes]
        split
            Name of data split corresponding to this dataset.
        input_data_key
            Name of data field to initialize in ``X_dict``
        task_name
            Name of task and corresponding label key in ``Y_dict``
        dataset_name
            Name of DictDataset to be initialized; See ``__init__`` above.

        Returns
        -------
        DictDataset
            Class initialized with single task and label corresponding to input data
        """
        return cls(
            name=dataset_name,
            split=split,
            X_dict={input_data_key: X_tensor},
            Y_dict={task_name: Y_tensor},
        )


def collate_dicts(batch: List[Batch]) -> Batch:
    """Combine many one-element dicts into a single many-element dict for both X and Y.

    Parameters
    ----------
    batch
        A list of (x_dict, y_dict) where the values of each are a single element

    Returns
    -------
    Batch
        A tuple of X_dict, Y_dict where the values of each are a merged list or tensor
    """
    X_batch: Dict[str, Any] = defaultdict(list)
    Y_batch: Dict[str, Any] = defaultdict(list)

    for x_dict, y_dict in batch:
        for field_name, value in x_dict.items():
            X_batch[field_name].append(value)
        for label_name, value in y_dict.items():
            Y_batch[label_name].append(value)

    for field_name, values in X_batch.items():
        # Only merge list of tensors
        if isinstance(values[0], Tensor):
            X_batch[field_name] = list_to_tensor(values)

    for label_name, values in Y_batch.items():
        Y_batch[label_name] = list_to_tensor(values)

    return dict(X_batch), dict(Y_batch)


class DictDataLoader(DataLoader):
    """A DataLoader that uses the appropriate collate_fn for a ``DictDataset``.

    Parameters
    ----------
    dataset
        A dataset to wrap
    collate_fn
        The collate function to use when combining multiple indexed examples for a
        single batch. Usually the default collate_dicts() method should be used, but
        it can be overriden if you want to use different collate logic.
    kwargs
        Keyword arguments to pass on to DataLoader.__init__()
    """

    def __init__(
        self,
        dataset: DictDataset,
        collate_fn: Callable[..., Any] = collate_dicts,
        **kwargs: Any,
    ) -> None:
        assert isinstance(dataset, DictDataset)
        super().__init__(dataset, collate_fn=collate_fn, **kwargs)



import urllib.parse


def connect_mongo(
    collection, database="report", server="localhost", username=None, password=None
):
    from pymongo import MongoClient

    if username is None:
        uri = "mongodb://%s/%s" % (server, database)
    else:
        username = urllib.parse.quote_plus(username)
        password = urllib.parse.quote_plus(password)
        uri = "mongodb://%s:%s@%s/%s?authMechanism=SCRAM-SHA-1" % (
            username,
            password,
            server,
            database,
        )
    # print('connecting to MongoDB: {}'.format(server))
    return MongoClient(uri)[database][collection]



def print_status(status):
    max_len_elapsed = max(max(len(row[4]) for row in status), 7)
    max_len_filename = max(max(len(row[5]) for row in status), 9)
    form = "| {:4} | {:25} | {:10} | {:13} | {} | {} |"
    print(
        "\n\n"
        + form.format(
            "Done",
            "Corpus name",
            "Num loaded",
            "Num processed",
            "Elapsed" + " " * (max_len_elapsed - 7),
            "File name" + " " * (max_len_filename - 9),
        )
    )
    print(
        form.format(
            "-" * 4,
            "-" * 25,
            "-" * 10,
            "-" * 13,
            "-" * max_len_elapsed,
            "-" * max_len_filename,
        )
    )
    for finish, name, num_loaded, num_sent, elapsed, filename in status:
        if not filename:
            filename = " " * max_len_filename
        else:
            filename += " " * (max_len_filename - len(filename))
        if not elapsed:
            elapsed = " " * max_len_elapsed
        else:
            elapsed += " " * (max_len_elapsed - len(elapsed))
        print(form.format(finish, name, num_loaded, num_sent, elapsed, filename))



import os
import logging
from ekorpkit import eKonf
from ekorpkit.pipelines.pipe import apply_pipeline
from .dataset import Dataset


log = logging.getLogger(__name__)


class FeatureSet(Dataset):
    """Feature class."""

    def __init__(self, **args):
        super().__init__(**args)

    @property
    def X_train(self):
        if self.train_data is not None:
            return self.train_data[self.COLUMN.X]
        else:
            return None

    @property
    def X_dev(self):
        if self.dev_data is not None:
            return self.dev_data[self.COLUMN.X]
        else:
            return None

    @property
    def X_test(self):
        if self.test_data is not None:
            return self.test_data[self.COLUMN.X]
        else:
            return None

    @property
    def y_train(self):
        if self.train_data is not None:
            return self.train_data[self.COLUMN.Y]
        else:
            return None

    @y_train.setter
    def y_train(self, data):
        self.train_data[self.COLUMN.Y] = data

    @property
    def y_dev(self):
        if self.dev_data is not None:
            return self.dev_data[self.COLUMN.Y]
        else:
            return None

    @y_dev.setter
    def y_dev(self, data):
        if self.dev_data is not None:
            self.dev_data[self.COLUMN.Y] = data

    @property
    def y_test(self):
        if self.test_data is not None:
            return self.test_data[self.COLUMN.Y]
        else:
            return None

    @y_test.setter
    def y_test(self, data):
        if self.test_data is not None:
            self.test_data[self.COLUMN.Y] = data

    @property
    def X(self):
        return self.data[self.COLUMN.X]

    @property
    def y(self):
        return self.data[self.COLUMN.Y]

    def persist(self):
        if not self._loaded:
            log.info(f"Dataset {self.name} is not loaded")
            return
        if self.summary_info is None:
            self.summarize()
        for split, data in self._splits.items():
            if data is None:
                continue
            data_file = self.data_files[split]
            eKonf.save_data(
                data,
                data_file,
                base_dir=self.data_dir,
                verbose=self.verbose,
            )
        if self.summary_info is not None:
            self.summary_info.save(info={"column_info": self.COLUMN.INFO})



import torch
import logging
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class Midas:
    def __init__(self, **args):
        args = eKonf.to_config(args)

        self.cuda_device = torch.device(
            f"cuda:{args.cuda_device}"
            if (torch.cuda.is_available() and not args.use_cpu)
            else "cpu"
        )

        # Initialize MiDaS depth model.
        # It remains resident in VRAM and likely takes around 2GB VRAM.
        # You could instead initialize it for each frame (and free it after each frame) to save VRAM.. but initializing it is slow.
        self.default_models = args.default_models

    def init_midas_depth_model(self, midas_model_type="dpt_large", optimize=True):
        import torchvision.transforms as T
        import cv2
        from midas.dpt_depth import DPTDepthModel
        from midas.midas_net import MidasNet
        from midas.midas_net_custom import MidasNet_small
        from midas.transforms import Resize, NormalizeImage, PrepareForNet

        midas_model = None
        net_w = None
        net_h = None
        resize_mode = None
        normalization = None

        log.info(f"Initializing MiDaS '{midas_model_type}' depth model...")
        # load network
        midas_model_path = self.default_models[midas_model_type]

        if midas_model_type == "dpt_large":  # DPT-Large
            midas_model = DPTDepthModel(
                path=midas_model_path,
                backbone="vitl16_384",
                non_negative=True,
            )
            net_w, net_h = 384, 384
            resize_mode = "minimal"
            normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif midas_model_type == "dpt_hybrid":  # DPT-Hybrid
            midas_model = DPTDepthModel(
                path=midas_model_path,
                backbone="vitb_rn50_384",
                non_negative=True,
            )
            net_w, net_h = 384, 384
            resize_mode = "minimal"
            normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif midas_model_type == "dpt_hybrid_nyu":  # DPT-Hybrid-NYU
            midas_model = DPTDepthModel(
                path=midas_model_path,
                backbone="vitb_rn50_384",
                non_negative=True,
            )
            net_w, net_h = 384, 384
            resize_mode = "minimal"
            normalization = NormalizeImage(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5])
        elif midas_model_type == "midas_v21":
            midas_model = MidasNet(midas_model_path, non_negative=True)
            net_w, net_h = 384, 384
            resize_mode = "upper_bound"
            normalization = NormalizeImage(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            )
        elif midas_model_type == "midas_v21_small":
            midas_model = MidasNet_small(
                midas_model_path,
                features=64,
                backbone="efficientnet_lite3",
                exportable=True,
                non_negative=True,
                blocks={"expand": True},
            )
            net_w, net_h = 256, 256
            resize_mode = "upper_bound"
            normalization = NormalizeImage(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            )
        else:
            print(f"midas_model_type '{midas_model_type}' not implemented")
            assert False

        midas_transform = T.Compose(
            [
                Resize(
                    net_w,
                    net_h,
                    resize_target=None,
                    keep_aspect_ratio=True,
                    ensure_multiple_of=32,
                    resize_method=resize_mode,
                    image_interpolation_method=cv2.INTER_CUBIC,
                ),
                normalization,
                PrepareForNet(),
            ]
        )

        midas_model.eval()

        if optimize == True:
            if self.cuda_device == torch.device("cuda"):
                midas_model = midas_model.to(memory_format=torch.channels_last)
                midas_model = midas_model.half()

        midas_model.to(self.cuda_device)

        log.info(f"MiDaS '{midas_model_type}' depth model initialized.")
        return midas_model, midas_transform, net_w, net_h, resize_mode, normalization



import logging
import os
import re
import pandas as pd
from glob import glob
from pathlib import Path, PosixPath, WindowsPath
from ekorpkit.utils.func import elapsed_timer


log = logging.getLogger(__name__)


def is_valid_regex(expr):
    try:
        if expr.startswith("r:"):
            expr = expr[2:]
        else:
            return False
        re.compile(expr)
        return True
    except re.error:
        return False


def glob_re(pattern, base_dir, recursive=False):
    if is_valid_regex(pattern):
        pattern = pattern[2:]
        pattern = re.compile(pattern)
        files = []
        if recursive:
            for (dirpath, dirnames, filenames) in os.walk(base_dir):
                files += [
                    os.path.join(dirpath, file)
                    for file in filenames
                    if pattern.search(file)
                ]
        else:
            files = [
                os.path.join(base_dir, file)
                for file in os.listdir(base_dir)
                if pattern.search(file)
            ]
    else:
        file = os.path.join(base_dir, pattern) if base_dir else pattern
        files = glob(file, recursive=recursive)
    return files


def get_filepaths(
    filename_patterns, base_dir=None, recursive=True, verbose=True, **kwargs
):
    if isinstance(filename_patterns, (PosixPath, WindowsPath)):
        filename_patterns = str(filename_patterns)
    if isinstance(filename_patterns, str):
        filename_patterns = [filename_patterns]
    filepaths = []
    for file in filename_patterns:
        filepath = os.path.join(base_dir, file) if base_dir else file
        if os.path.exists(filepath):
            if Path(filepath).is_file():
                filepaths.append(filepath)
        else:
            filepaths += glob_re(file, base_dir, recursive=recursive)
    filepaths = [fp for fp in filepaths if Path(fp).is_file()]
    if verbose:
        log.info(f"Processing [{len(filepaths)}] files from {filename_patterns}")

    return filepaths


def get_files_from_archive(archive_path, filetype=None):
    import tarfile
    from zipfile import ZipFile

    if ".tar.gz" in archive_path:
        log.info(f"::Extracting files from {archive_path} with tar.gz")
        archive_handle = tarfile.open(archive_path, "r:gz")
        files = [
            (file, file.name) for file in archive_handle.getmembers() if file.isfile()
        ]
        open_func = archive_handle.extractfile
    elif ".tar.bz2" in archive_path:
        log.info(f"::Extracting files from {archive_path} with tar.bz2")
        archive_handle = tarfile.open(archive_path, "r:bz2")
        files = [
            (file, file.name) for file in archive_handle.getmembers() if file.isfile()
        ]
        open_func = archive_handle.extractfile
    elif ".zip" in archive_path:
        log.info(f"::Extracting files from {archive_path} with zip")
        archive_handle = ZipFile(archive_path)
        files = [
            (file, file.encode("cp437").decode("euc-kr"))
            for file in archive_handle.namelist()
        ]
        open_func = archive_handle.open
    else:
        # print(f'::{archive_path} is not archive, use generic method')
        files = [(archive_path, os.path.basename(archive_path))]
        archive_handle = None
        open_func = None
    if filetype:
        files = [file for file in files if filetype in file[1]]

    return files, archive_handle, open_func


def is_dataframe(data):
    return isinstance(data, pd.DataFrame)


def concat_data(
    data,
    columns=None,
    add_key_as_name=False,
    name_column="_name_",
    ignore_index=True,
    verbose=False,
    **kwargs,
):
    if isinstance(data, dict):
        log.info(f"Concatenating {len(data)} dataframes")
        dfs = []
        for df_name in data:
            df_each = data[df_name]
            if isinstance(columns, list):
                _columns = [c for c in columns if c in df_each.columns]
                df_each = df_each[_columns]
            if add_key_as_name:
                df_each[name_column] = df_name
            dfs.append(df_each)
        if len(dfs) > 0:
            return pd.concat(dfs, ignore_index=ignore_index)
        else:
            return None
    elif isinstance(data, list):
        log.info(f"Concatenating {len(data)} dataframes")
        if len(data) > 0:
            return pd.concat(data, ignore_index=ignore_index)
        else:
            return None
    else:
        log.warning("Warning: data is not a dict")
        return data


def load_data(filename, base_dir=None, verbose=False, **kwargs):
    concatenate = kwargs.pop("concatenate", False)
    ignore_index = kwargs.pop("ignore_index", False)

    if base_dir:
        filepaths = get_filepaths(filename, base_dir)
    else:
        filepaths = get_filepaths(filename)
    log.info(f"Loading {len(filepaths)} dataframes from {filepaths}")

    data = {
        os.path.basename(f): load_dataframe(f, verbose=verbose, **kwargs)
        for f in filepaths
    }
    data = {k: v for k, v in data.items() if v is not None}
    if len(data) == 1:
        return list(data.values())[0]
    elif len(filepaths) > 1:
        if concatenate:
            return pd.concat(data.values(), ignore_index=ignore_index)
        else:
            return data
    else:
        log.warning(f"No files found for {filename}")
        return None


def load_dataframe(
    filename,
    base_dir=None,
    columns=None,
    index_col=None,
    verbose=False,
    **kwargs,
):
    dtype = kwargs.pop("dtype", None)
    if isinstance(dtype, list):
        dtype = {k: "str" for k in dtype}
    parse_dates = kwargs.pop("parse_dates", False)

    filetype = kwargs.pop("filetype", None) or "parquet"
    fileinfo = os.path.splitext(filename)
    filename = fileinfo[0]
    filetype = fileinfo[1] if len(fileinfo) > 1 else filetype
    filetype = "." + filetype.replace(".", "")
    filename = f"{filename}{filetype}"
    if base_dir is not None:
        filepath = os.path.join(base_dir, filename)
    else:
        filepath = filename

    if not os.path.exists(filepath):
        log.warning(f"File {filepath} does not exist")
        return None
    log.info(f"Loading data from {filepath}")
    with elapsed_timer(format_time=True) as elapsed:
        if "csv" in filetype or "tsv" in filetype:
            delimiter = kwargs.pop("delimiter", "\t") if "tsv" in filetype else None
            data = pd.read_csv(
                filepath,
                index_col=index_col,
                dtype=dtype,
                parse_dates=parse_dates,
                delimiter=delimiter,
            )
        elif "parquet" in filetype:
            engine = kwargs.pop("engine", "pyarrow")
            data = pd.read_parquet(filepath, engine=engine)
        else:
            raise ValueError("filetype must be .csv or .parquet")
        if isinstance(columns, list):
            columns = [c for c in columns if c in data.columns]
            data = data[columns]
        if verbose:
            log.info(" >> elapsed time to load data: {}".format(elapsed()))
    return data


def save_data(
    data,
    filename,
    base_dir=None,
    columns=None,
    index=False,
    filetype="parquet",
    suffix=None,
    verbose=False,
    **kwargs,
):
    fileinfo = os.path.splitext(filename)
    filename = fileinfo[0]
    filetype = fileinfo[1] if len(fileinfo) > 1 else filetype
    filetype = "." + filetype.replace(".", "")
    if suffix is not None:
        filename = f"{filename}-{suffix}{filetype}"
    else:
        filename = f"{filename}{filetype}"
    if base_dir is not None:
        filepath = os.path.join(base_dir, filename)
    else:
        filepath = filename
    base_dir = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    os.makedirs(base_dir, exist_ok=True)

    if isinstance(data, dict):
        for k, v in data.items():
            save_data(
                v,
                filename,
                base_dir=base_dir,
                columns=columns,
                index=index,
                filetype=filetype,
                suffix=k,
                verbose=verbose,
                **kwargs,
            )
    elif is_dataframe(data):
        log.info(f"Saving dataframe to {filepath}")
        if isinstance(columns, list):
            columns = [c for c in columns if c in data.columns]
            data = data[columns]
        with elapsed_timer(format_time=True) as elapsed:
            if "csv" in filetype or "tsv" in filetype:
                data.to_csv(filepath, index=index)
            elif "parquet" in filetype:
                compression = kwargs.get("compression", "gzip")
                engine = kwargs.get("engine", "pyarrow")
                data.to_parquet(filepath, compression=compression, engine=engine)
            else:
                raise ValueError("filetype must be .csv or .parquet")
            if verbose:
                log.info(" >> elapsed time to save data: {}".format(elapsed()))
    else:
        raise ValueError(f"Unsupported data type: {type(data)}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Korean Sentence Splitter
# Split Korean text into sentences using heuristic algorithm.
#
# Copyright (C) 2021 Hyun-Woong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from typing import Optional, Any


class Node:
    def __init__(self, key, val):
        self.key = key
        self.val = val
        self.next = None
        self.prev = None


class LRUCache:
    def __init__(self, max_size: int):
        self.head = Node("head", "head")
        self.tail = Node("tail", "tail")
        self.head.next = self.tail
        self.tail.prev = self.head
        self.max_size = max_size
        self.curr_size = 0
        self.dic = {}  # hashmap

    def removeNode(self, node):
        """node의 앞 뒤를 서로 연결"""
        node_after, node_before = node.next, node.prev
        node_before.next, node_after.prev = node_after, node_before

    def moveToEnd(self, node):
        prev_node_tail = self.tail.prev
        prev_node_tail.next = node
        node.prev = prev_node_tail
        node.next = self.tail
        self.tail.prev = node

    def get(self, key: Any) -> Optional[Any]:
        # 해쉬맵에서 못 찾으면 None
        if key not in self.dic:
            return None

        node = self.dic[key]
        node_val = node.val
        # 가장 최근에 찾은 노드는 연결 리스트의 맨 뒤(앞에서부터 삭제)로 이동
        self.removeNode(node)
        self.moveToEnd(node)
        return node_val

    def put(self, key: Any, new_val: Any) -> None:
        if key in self.dic:
            node = self.dic[key]
            node.val = new_val
            self.removeNode(node)
            self.moveToEnd(node)
        else:
            if self.curr_size >= self.max_size:
                # 연결리스트 길이가 최대 길이를 넘어갈 경우 앞에서부터 삭제
                old_node = self.head.next
                del self.dic[old_node.key]
                self.removeNode(old_node)
                self.curr_size -= 1

            new_node = Node(key, new_val)
            self.dic[key] = new_node
            self.moveToEnd(new_node)
            self.curr_size += 1



import logging
import os
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional

import torch
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm

from snorkel.classification.data import DictDataLoader  # noqa: F401
from snorkel.classification.multitask_classifier import (
    ClassifierConfig,
    MultitaskClassifier,
)
from snorkel.types import Config
from snorkel.utils.config_utils import merge_config
from snorkel.utils.lr_schedulers import LRSchedulerConfig
from snorkel.utils.optimizers import OptimizerConfig

from .loggers import (
    Checkpointer,
    CheckpointerConfig,
    LogManager,
    LogManagerConfig,
    LogWriter,
    LogWriterConfig,
    TensorBoardWriter,
)
from .schedulers import batch_schedulers

Metrics = Dict[str, float]


class TrainerConfig(Config):
    """Settings for the Trainer.

    Parameters
    ----------
    seed
        A random seed to set before training; if None, no seed is set
    n_epochs
        The number of epochs to train
    lr
        Base learning rate (will also be affected by lr_scheduler choice and settings)
    l2
        L2 regularization coefficient (weight decay)
    grad_clip
        The value that the gradient norm will be clipped to if it exceeds it
    train_split
        The name of the split to use as the training set
    valid_split
        The name of the split to use as the validation set
    test_split
        The name of the split to use as the test set
    progress_bar
        If True, print a tqdm progress bar during training
    model_config
        Settings for the MultitaskClassifier
    log_manager_config
        Settings for the LogManager
    checkpointing
        If True, use a Checkpointer to save the best model during training
    checkpointer_config
        Settings for the Checkpointer
    logging
        If True, log metrics (to file or Tensorboard) during training
    log_writer
        The type of LogWriter to use (one of ["json", "tensorboard"])
    log_writer_config
        Settings for the LogWriter
    optimizer
        Which optimizer to use (one of ["sgd", "adam", "adamax"])
    optimizer_config
        Settings for the optimizer
    lr_scheduler
        Which lr_scheduler to use (one of ["constant", "linear", "exponential", "step"])
    lr_scheduler_config
        Settings for the LRScheduler
    batch_scheduler
        Which batch scheduler to use (in what order batches will be drawn from multiple
        tasks)
    """

    seed: Optional[int] = None
    n_epochs: int = 1
    lr: float = 0.01
    l2: float = 0.0
    grad_clip: float = 1.0
    train_split: str = "train"
    valid_split: str = "valid"
    test_split: str = "test"
    progress_bar: bool = True
    model_config: ClassifierConfig = ClassifierConfig()  # type:ignore
    log_manager_config: LogManagerConfig = LogManagerConfig()  # type:ignore
    checkpointing: bool = False
    checkpointer_config: CheckpointerConfig = CheckpointerConfig()  # type:ignore
    logging: bool = False
    log_writer: str = "tensorboard"
    log_writer_config: LogWriterConfig = LogWriterConfig()  # type:ignore
    optimizer: str = "adam"
    optimizer_config: OptimizerConfig = OptimizerConfig()  # type:ignore
    lr_scheduler: str = "constant"
    lr_scheduler_config: LRSchedulerConfig = LRSchedulerConfig()  # type:ignore
    batch_scheduler: str = "shuffled"


class Trainer:
    """A class for training a MultitaskClassifier.

    Parameters
    ----------
    name
        An optional name for this trainer object
    kwargs
        Settings to be merged into the default Trainer config dict

    Attributes
    ----------
    name
        See above
    config
        The config dict with the settings for the Trainer
    checkpointer
        Saves the best model seen during training
    log_manager
        Identifies when its time to log or evaluate on the valid set
    log_writer
        Writes training statistics to file or TensorBoard
    optimizer
        Updates model weights based on the loss
    lr_scheduler
        Adjusts the learning rate over the course of training
    batch_scheduler
        Returns batches from the DataLoaders in a particular order for training
    """

    def __init__(self, name: Optional[str] = None, **kwargs: Any) -> None:
        self.config: TrainerConfig = merge_config(  # type:ignore
            TrainerConfig(), kwargs  # type:ignore
        )
        self.name = name if name is not None else type(self).__name__

    def fit(
        self, model: MultitaskClassifier, dataloaders: List["DictDataLoader"]
    ) -> None:
        """Train a MultitaskClassifier.

        Parameters
        ----------
        model
            The model to train
        dataloaders
            A list of DataLoaders. These will split into train, valid, and test splits
            based on the ``split`` attribute of the DataLoaders.
        """
        self._check_dataloaders(dataloaders)

        # Identify the dataloaders to train on
        train_dataloaders = [
            dl
            for dl in dataloaders
            if dl.dataset.split == self.config.train_split  # type: ignore
        ]

        # Calculate the total number of batches per epoch
        self.n_batches_per_epoch = sum(
            [len(dataloader) for dataloader in train_dataloaders]
        )

        # Set training helpers
        self._set_log_writer()
        self._set_checkpointer()
        self._set_log_manager()
        self._set_optimizer(model)
        self._set_lr_scheduler()
        self._set_batch_scheduler()

        # Set to training mode
        model.train()

        logging.info("Start training...")

        self.metrics: Dict[str, float] = dict()
        self._reset_losses()

        for epoch_num in range(self.config.n_epochs):
            batches = tqdm(
                enumerate(self.batch_scheduler.get_batches(train_dataloaders)),
                total=self.n_batches_per_epoch,
                disable=(not self.config.progress_bar),
                desc=f"Epoch {epoch_num}:",
            )
            for batch_num, (batch, dataloader) in batches:
                X_dict, Y_dict = batch

                total_batch_num = epoch_num * self.n_batches_per_epoch + batch_num
                batch_size = len(next(iter(Y_dict.values())))

                # Set gradients of all model parameters to zero
                self.optimizer.zero_grad()

                # Perform forward pass and calcualte the loss and count
                loss_dict, count_dict = model.calculate_loss(X_dict, Y_dict)

                # Update running loss and count
                for task_name in loss_dict.keys():
                    identifier = "/".join(
                        [
                            task_name,
                            dataloader.dataset.name,
                            dataloader.dataset.split,
                            "loss",
                        ]
                    )
                    self.running_losses[identifier] += (
                        loss_dict[task_name].item() * count_dict[task_name]
                    )
                    self.running_counts[task_name] += count_dict[task_name]

                # Skip the backward pass if no loss is calcuated
                if not loss_dict:
                    continue

                # Calculate the average loss
                loss = torch.stack(list(loss_dict.values())).sum()

                # Perform backward pass to calculate gradients
                loss.backward()

                # Clip gradient norm
                if self.config.grad_clip:
                    torch.nn.utils.clip_grad_norm_(
                        model.parameters(), self.config.grad_clip
                    )

                # Update the parameters
                self.optimizer.step()

                # Update lr using lr scheduler
                self._update_lr_scheduler(total_batch_num)

                # Update metrics
                self.metrics.update(self._logging(model, dataloaders, batch_size))

                batches.set_postfix(self.metrics)

        model = self.log_manager.cleanup(model)

    def _check_dataloaders(self, dataloaders: List["DictDataLoader"]) -> None:
        """Validate the dataloader splits."""
        train_split = self.config.train_split
        valid_split = self.config.valid_split
        test_split = self.config.test_split

        all_splits = [train_split, valid_split, test_split]
        if not all(d.dataset.split in all_splits for d in dataloaders):  # type: ignore
            raise ValueError(f"Dataloader splits must be one of {all_splits}")

        if not any(d.dataset.split == train_split for d in dataloaders):  # type: ignore
            raise ValueError(
                f"Cannot find any dataloaders with split matching train split: "
                f"{self.config.train_split}."
            )

    def _set_log_writer(self) -> None:
        self.log_writer: Optional[LogWriter] = None
        if self.config.logging:
            if self.config.log_writer == "json":
                self.log_writer = LogWriter(**self.config.log_writer_config._asdict())
            elif self.config.log_writer == "tensorboard":
                self.log_writer = TensorBoardWriter(
                    **self.config.log_writer_config._asdict()
                )
            else:
                raise ValueError(
                    f"Unrecognized writer option: {self.config.log_writer}"
                )

    def _set_checkpointer(self) -> None:
        self.checkpointer: Optional[Checkpointer]

        if self.config.checkpointing:
            checkpointer_config = self.config.checkpointer_config
            evaluation_freq = self.config.log_manager_config.evaluation_freq
            counter_unit = self.config.log_manager_config.counter_unit
            self.checkpointer = Checkpointer(
                counter_unit, evaluation_freq, **checkpointer_config._asdict()
            )
        else:
            self.checkpointer = None

    def _set_log_manager(self) -> None:
        self.log_manager = LogManager(
            self.n_batches_per_epoch,
            log_writer=self.log_writer,
            checkpointer=self.checkpointer,
            **self.config.log_manager_config._asdict(),
        )

    def _set_optimizer(self, model: nn.Module) -> None:
        optimizer_config = self.config.optimizer_config
        optimizer_name = self.config.optimizer

        parameters = filter(lambda p: p.requires_grad, model.parameters())

        optimizer: optim.Optimizer  # type: ignore

        if optimizer_name == "sgd":
            optimizer = optim.SGD(  # type: ignore
                parameters,
                lr=self.config.lr,
                weight_decay=self.config.l2,
                **optimizer_config.sgd_config._asdict(),
            )
        elif optimizer_name == "adam":
            optimizer = optim.Adam(
                parameters,
                lr=self.config.lr,
                weight_decay=self.config.l2,
                **optimizer_config.adam_config._asdict(),
            )
        elif optimizer_name == "adamax":
            optimizer = optim.Adamax(  # type: ignore
                parameters,
                lr=self.config.lr,
                weight_decay=self.config.l2,
                **optimizer_config.adamax_config._asdict(),
            )
        else:
            raise ValueError(f"Unrecognized optimizer option '{optimizer_name}'")

        logging.info(f"Using optimizer {optimizer}")

        self.optimizer = optimizer

    def _set_lr_scheduler(self) -> None:
        # Set warmup scheduler
        self._set_warmup_scheduler()

        # Set lr scheduler
        lr_scheduler_name = self.config.lr_scheduler
        lr_scheduler_config = self.config.lr_scheduler_config
        lr_scheduler: Optional[optim.lr_scheduler._LRScheduler]

        if lr_scheduler_name == "constant":
            lr_scheduler = None
        elif lr_scheduler_name == "linear":
            total_steps = self.n_batches_per_epoch * self.config.n_epochs
            linear_decay_func = lambda x: (total_steps - self.warmup_steps - x) / (
                total_steps - self.warmup_steps
            )
            lr_scheduler = optim.lr_scheduler.LambdaLR(  # type: ignore
                self.optimizer, linear_decay_func
            )
        elif lr_scheduler_name == "exponential":
            lr_scheduler = optim.lr_scheduler.ExponentialLR(
                self.optimizer, **lr_scheduler_config.exponential_config._asdict()
            )
        elif lr_scheduler_name == "step":
            lr_scheduler = optim.lr_scheduler.StepLR(
                self.optimizer, **lr_scheduler_config.step_config._asdict()
            )
        else:
            raise ValueError(f"Unrecognized lr scheduler option '{lr_scheduler_name}'")

        self.lr_scheduler = lr_scheduler

    def _set_warmup_scheduler(self) -> None:
        warmup_scheduler: Optional[optim.lr_scheduler.LambdaLR]

        if self.config.lr_scheduler_config.warmup_steps:
            warmup_steps = self.config.lr_scheduler_config.warmup_steps
            if warmup_steps < 0:
                raise ValueError("warmup_steps much greater or equal than 0.")
            warmup_unit = self.config.lr_scheduler_config.warmup_unit
            if warmup_unit == "epochs":
                self.warmup_steps = int(warmup_steps * self.n_batches_per_epoch)
            elif warmup_unit == "batches":
                self.warmup_steps = int(warmup_steps)
            else:
                raise ValueError(
                    f"warmup_unit must be 'batches' or 'epochs', but {warmup_unit} found."
                )
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(  # type: ignore
                self.optimizer, linear_warmup_func
            )
            logging.info(f"Warmup {self.warmup_steps} batches.")
        elif self.config.lr_scheduler_config.warmup_percentage:
            warmup_percentage = self.config.lr_scheduler_config.warmup_percentage
            self.warmup_steps = int(
                warmup_percentage * self.config.n_epochs * self.n_batches_per_epoch
            )
            linear_warmup_func = lambda x: x / self.warmup_steps
            warmup_scheduler = optim.lr_scheduler.LambdaLR(  # type: ignore
                self.optimizer, linear_warmup_func
            )
            logging.info(f"Warmup {self.warmup_steps} batches.")
        else:
            warmup_scheduler = None
            self.warmup_steps = 0

        self.warmup_scheduler = warmup_scheduler

    def _update_lr_scheduler(self, step: int) -> None:
        if self.warmup_scheduler and step < self.warmup_steps:
            self.warmup_scheduler.step()  # type: ignore
        elif self.lr_scheduler is not None:
            self.lr_scheduler.step()  # type: ignore
            min_lr = self.config.lr_scheduler_config.min_lr
            if min_lr and self.optimizer.param_groups[0]["lr"] < min_lr:
                self.optimizer.param_groups[0]["lr"] = min_lr

    def _set_batch_scheduler(self) -> None:
        scheduler_class = batch_schedulers.get(self.config.batch_scheduler)
        if not scheduler_class:
            raise ValueError(f"Unrecognized batch scheduler option '{scheduler_class}'")

        self.batch_scheduler = scheduler_class()  # type: ignore

    def _evaluate(
        self,
        model: MultitaskClassifier,
        dataloaders: List["DictDataLoader"],
        split: str,
    ) -> Metrics:
        """Evalute the current quality of the model on data for the requested split."""
        loaders = [d for d in dataloaders if d.dataset.split in split]  # type: ignore
        return model.score(loaders)

    def _logging(
        self,
        model: MultitaskClassifier,
        dataloaders: List["DictDataLoader"],
        batch_size: int,
    ) -> Metrics:
        """Log and checkpoint if it is time to do so."""

        # Switch to eval mode for evaluation
        model.eval()

        self.log_manager.update(batch_size)

        # Log the loss and lr
        metric_dict: Metrics = dict()
        metric_dict.update(self._aggregate_losses())

        # Evaluate the model and log the metric
        if self.log_manager.trigger_evaluation():

            # Log metrics
            metric_dict.update(
                self._evaluate(model, dataloaders, self.config.valid_split)
            )
            self._log_metrics(metric_dict)
            self._reset_losses()

        # Checkpoint the model
        if self.log_manager.trigger_checkpointing():
            self._checkpoint_model(model, metric_dict)
            self._reset_losses()

        # Switch back to train mode
        model.train()
        return metric_dict

    def _log_metrics(self, metric_dict: Metrics) -> None:
        if self.log_writer is not None:
            for metric_name, metric_value in metric_dict.items():
                self.log_writer.add_scalar(
                    metric_name, metric_value, self.log_manager.point_total
                )

    def _checkpoint_model(
        self, model: MultitaskClassifier, metric_dict: Metrics
    ) -> None:
        """Save the current model."""
        if self.checkpointer is not None:
            self.checkpointer.checkpoint(
                self.log_manager.unit_total, model, metric_dict
            )

    def _aggregate_losses(self) -> Metrics:
        """Calculate the task specific loss, average micro loss and learning rate."""

        metric_dict = dict()

        # Log task specific loss
        self.running_losses: DefaultDict[str, float]
        self.running_counts: DefaultDict[str, float]
        for identifier in self.running_losses.keys():
            if self.running_counts[identifier] > 0:
                metric_dict[identifier] = (
                    self.running_losses[identifier] / self.running_counts[identifier]
                )

        # Calculate average micro loss
        total_loss = sum(self.running_losses.values())
        total_count = sum(self.running_counts.values())
        if total_count > 0:
            metric_dict["model/all/train/loss"] = total_loss / total_count

        # Log the learning rate
        metric_dict["model/all/train/lr"] = self.optimizer.param_groups[0]["lr"]

        return metric_dict

    def _reset_losses(self) -> None:
        """Reset the loss counters."""
        self.running_losses = defaultdict(float)
        self.running_counts = defaultdict(int)

    def save(self, trainer_path: str) -> None:
        """Save the trainer config to the specified file path in json format.

        Parameters
        ----------
        trainer_path
            The path where trainer config and optimizer state should be saved.
        """

        head, tail = os.path.split(trainer_path)

        if not os.path.exists(head):
            os.makedirs(os.path.dirname(head))
        try:
            torch.save(
                {
                    "trainer_config": self.config._asdict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                trainer_path,
            )
        except BaseException:  # pragma: no cover
            logging.warning("Saving failed... continuing anyway.")

        logging.info(f"[{self.name}] Trainer config saved in {trainer_path}")

    def load(self, trainer_path: str, model: Optional[MultitaskClassifier]) -> None:
        """Load trainer config and optimizer state from the specified json file path to the trainer object. The optimizer state is stored, too. However, it only makes sense if loaded with the correct model again.

        Parameters
        ----------
        trainer_path
            The path to the saved trainer config to be loaded
        model
            MultitaskClassifier for which the optimizer has been set. Parameters of optimizer must fit to model parameters. This model
            shall be the model which was fit by the stored Trainer.

        Example
        -------
        Saving model and corresponding trainer:
        >>> model.save('./my_saved_model_file') # doctest: +SKIP
        >>> trainer.save('./my_saved_trainer_file') # doctest: +SKIP
        Now we can resume training and load the saved model and trainer into new model and trainer objects:
        >>> new_model.load('./my_saved_model_file') # doctest: +SKIP
        >>> new_trainer.load('./my_saved_trainer_file', model=new_model) # doctest: +SKIP
        >>> new_trainer.fit(...) # doctest: +SKIP
        """

        try:
            saved_state = torch.load(trainer_path)
        except BaseException:
            if not os.path.exists(trainer_path):
                logging.error("Loading failed... Trainer config does not exist.")
            else:
                logging.error(
                    f"Loading failed... Cannot load trainer config from {trainer_path}"
                )
            raise

        self.config = TrainerConfig(**saved_state["trainer_config"])
        logging.info(f"[{self.name}] Trainer config loaded from {trainer_path}")

        if model is not None:
            try:
                self._set_optimizer(model)
                self.optimizer.load_state_dict(saved_state["optimizer_state_dict"])
                logging.info(f"[{self.name}] Optimizer loaded from {trainer_path}")
            except BaseException:
                logging.error(
                    "Loading the optimizer for your model failed. Optimizer state NOT loaded."
                )



from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

from ...classification.utils import collect_flow_outputs_by_suffix


class SliceCombinerModule(nn.Module):
    """A module for combining the weighted representations learned by slices.

    Intended for use with the MultitaskClassifier including:
        * Indicator operations
        * Prediction operations
        * Prediction transform features

    NOTE: This module currently only handles binary labels.

    Parameters
    ----------
    slice_ind_key
        Suffix of operation corresponding to the slice indicator heads
    slice_pred_key
        Suffix of operation corresponding to the slice predictor heads
    slice_pred_feat_key
        Suffix of operation corresponding to the slice predictor features heads
    temperature
        Temperature constant for scaling the weighting between indicator prediction
        and predictor confidences: SoftMax(indicator_pred * predictor_confidence / tau)

    Attributes
    ----------
    slice_ind_key
        See above
    slice_pred_key
        See above
    slice_pred_feat_key
        See above
    """

    def __init__(
        self,
        slice_ind_key: str = "_ind_head",
        slice_pred_key: str = "_pred_head",
        slice_pred_feat_key: str = "_pred_transform",
        temperature: float = 1.0,
    ) -> None:
        super().__init__()

        self.slice_ind_key = slice_ind_key
        self.slice_pred_key = slice_pred_key
        self.slice_pred_feat_key = slice_pred_feat_key
        self.temperature = temperature

    def forward(  # type:ignore
        self, output_dict: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        """Reweight and combine predictor representations given output dict.

        Parameters
        ----------
        output_dict
            A dict of data fields containing operation outputs from indicator head,
            predictor head, and predictor transform (corresponding to slice_ind_key,
            slice_pred_key, slice_pred_feat_key, respectively).

            NOTE: The output_dict outputs for the ind/pred heads must be raw logits.

        Returns
        -------
        torch.Tensor
            The reweighted predictor representation
        """

        # Concatenate indicator head predictions into tensor [batch_size x num_slices]
        indicator_outputs = collect_flow_outputs_by_suffix(
            output_dict, self.slice_ind_key
        )
        # Indicator heads are always binary (negative class = 0, positive class = 1)
        indicator_preds = torch.cat(
            [
                F.softmax(output, dim=1)[:, 1].unsqueeze(1)
                for output in indicator_outputs
            ],
            dim=-1,
        )

        # Concatenate predictor head confidences into tensor [batch_size x num_slices]
        predictor_outputs = collect_flow_outputs_by_suffix(
            output_dict, self.slice_pred_key
        )

        if predictor_outputs[0].shape[1] > 2:
            raise NotImplementedError(
                "SliceCombiner does not support more than 2 classes yet."
            )
        elif predictor_outputs[0].shape[1] < 2:
            raise NotImplementedError(
                "SliceCombiner currently requires output shape [..., 2] for predictor heads."
            )

        predictor_confidences = torch.cat(
            [
                # Compute the "confidence" using score of the positive class
                F.softmax(output, dim=1)[:, 1].unsqueeze(1)
                for output in predictor_outputs
            ],
            dim=-1,
        )

        # Concatenate each predictor feature (to be combined into reweighted
        # representation) into [batch_size x 1 x feat_dim] tensor
        predictor_feat_outputs = collect_flow_outputs_by_suffix(
            output_dict, self.slice_pred_feat_key
        )
        slice_representations = torch.stack(predictor_feat_outputs, dim=1)

        # Attention weights used to combine each of the slice_representations
        # incorporates the indicator (whether we are in the slice or not) and
        # predictor (confidence of a learned slice head)
        attention_weights = F.softmax(
            indicator_preds * predictor_confidences / self.temperature, dim=1
        )

        # Match dims [batch_size x num_slices x feat_dim] of slice_representations
        attention_weights = attention_weights.unsqueeze(-1).expand(
            [-1, -1, slice_representations.size(-1)]
        )

        # Reweight representations with weighted sum across slices
        reweighted_rep = torch.sum(attention_weights * slice_representations, dim=1)
        return reweighted_rep



import io
import requests
from typing import Optional
from .scheme_client import SchemeClient


RECOVERABLE_SERVER_ERROR_CODES = (502, 503, 504)


class RecoverableServerError(requests.exceptions.HTTPError):
    """
    Server returned one of `RECOVERABLE_SERVER_ERROR_CODES`.
    """


def session_with_backoff() -> requests.Session:
    """
    We ran into an issue where http requests to s3 were timing out,
    possibly because we were making too many requests too quickly.
    This helper function returns a requests session that has retry-with-backoff
    built in. See
    <https://stackoverflow.com/questions/23267409/how-to-implement-retry-mechanism-into-python-requests-library>.
    """
    from requests.packages.urllib3.util.retry import Retry
    from requests.adapters import HTTPAdapter

    session = requests.Session()
    retries = Retry(
        total=5, backoff_factor=1, status_forcelist=RECOVERABLE_SERVER_ERROR_CODES
    )
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session


class HttpClient(SchemeClient):
    scheme = ("http", "https")
    recoverable_errors = SchemeClient.recoverable_errors + (RecoverableServerError,)

    def __init__(self, resource: str) -> None:
        super().__init__(resource)
        self._head_response = None

    @property
    def head_response(self):
        if self._head_response is None:
            with session_with_backoff() as session:
                response = session.head(self.resource, allow_redirects=True)
            self.validate_response(response)
            self._head_response = response
            return self._head_response
        else:
            return self._head_response

    def get_etag(self) -> Optional[str]:
        return self.head_response.headers.get("ETag")

    def get_size(self) -> Optional[int]:
        content_length = self.head_response.headers.get("Content-Length")
        return None if content_length is None else int(content_length)

    def get_resource(self, temp_file: io.BufferedWriter) -> None:
        with session_with_backoff() as session:
            response = session.get(self.resource, stream=True)
            self.validate_response(response)
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    temp_file.write(chunk)

    def validate_response(self, response):
        if response.status_code == 404:
            raise FileNotFoundError(self.resource)
        if response.status_code in RECOVERABLE_SERVER_ERROR_CODES:
            raise RecoverableServerError(response=response)
        response.raise_for_status()



# original source:
# https://gist.githubusercontent.com/Jmkernes/01da3b560eb12218119f00a0969787e8/raw/2cc222706bdd8b15f2807b6fbd2c0612d7ce17ea/sentence_piece_trainer.py
import re
import collections
import numpy as np
from scipy.special import digamma


# To efficiently determine the next possible words
# We need a Trie data structure
class Trie:
    def __init__(self, end_symbol="<END>"):
        self.root = {}
        self.end_symbol = end_symbol

    def add(self, word, value):
        node = self.root
        for ch in word:
            if ch not in node:
                node[ch] = {}
            node = node[ch]
        node[self.end_symbol] = value

    def get_value(self, word):
        node = self.root
        for ch in word:
            if ch not in node:
                return 0
            node = node[ch]
        if self.end_symbol not in node:
            return 0
        return node[self.end_symbol]

    def set_value(self, word, value):
        node = self.root
        for ch in word:
            if ch not in node:
                raise ValueError("word not in trie")
            node = node[ch]
        if self.end_symbol not in node:
            raise ValueError("word not in trie")
        node[self.end_symbol] = value


class SentencePieceTokenizer:
    def __init__(
        self,
        initial_vocab_size=2000,
        vocab_size=1000,
        percent_to_prune=0.2,
        whitespace_token="▁",
        lowercase=True,
    ):
        self.trie = None
        self.maxlen = None

        self.initial_vocab_size = initial_vocab_size
        self.vocab_size = vocab_size
        self.percent_to_prune = percent_to_prune
        self.whitespace_token = whitespace_token
        self.lowercase = lowercase

    def _initialize_trie(self, tokens):
        trie = Trie()
        norm = sum(list(tokens.values()))
        logsum = digamma(norm)

        maxlen = 0
        for tok, val in tokens.items():
            trie.add(tok, digamma(val) - logsum)
            maxlen = max(maxlen, len(tok))

        return trie, maxlen

    def pre_tokenize(self, text):
        if self.lowercase:
            text = text.lower()
        text = re.sub(r"\s+", self.whitespace_token, text)
        return text

    def initialize_vocab(self, texts, initial_vocab_size=None):
        if initial_vocab_size is None:
            initial_vocab_size = self.initial_vocab_size
        text = self.pre_tokenize(" ".join(texts))
        word_freqs = collections.Counter(text.split(self.whitespace_token))
        sorted_subwords, characters = self.initialize_subwords(word_freqs)
        tokens = (
            list(characters.items())
            + sorted_subwords[: self.initial_vocab_size - len(characters)]
        )
        tokens = {token: freq for token, freq in tokens}
        tokens = collections.Counter(tokens)
        return text, tokens, characters

    def initialize_subwords(self, word_freqs, verbose=True):
        character_freqs = collections.defaultdict(int)
        subwords_freqs = collections.defaultdict(int)
        for word, freq in word_freqs.items():
            word = self.whitespace_token + word
            for i in range(len(word)):
                character_freqs[word[i]] += freq
                # Loop through the subwords of length at least 2
                for j in range(i + 2, len(word) + 1):
                    subwords_freqs[word[i:j]] += freq

        # Sort subwords by frequency
        sorted_subwords = sorted(
            subwords_freqs.items(), key=lambda x: x[1], reverse=True
        )
        if verbose:
            print("Top 10 subwords: {}".format(sorted_subwords[:10]))
        return sorted_subwords, character_freqs

    def forward_step(self, text, trie):
        N = len(text)

        # d[i] contains the maximum log_prob of any tokenization
        # of text[:i], initialized to 0 (i.e. log(0)=-infty)
        d = [-np.inf] * (N + 1)

        # p[i] (stands for parent) contains the number of characters of
        # the final token in the most likely sequence that ends at index i
        p = [None] * (N + 1)
        d[0] = 0

        for i in range(1, N + 1):

            # find all possible final words. Have to look back
            # a distance set by the length of the longest token
            for j in range(max(i - self.maxlen, 0), i):

                final_token = text[j:i]
                final_value = trie.get_value(final_token)

                # if the current ending word has a higher log-probability,
                # save that value and store the word (i.e. # chars to backtrack)
                if final_value and d[j] + final_value > d[i]:
                    d[i] = d[j] + final_value
                    p[i] = len(final_token)
            if p[i] is None:
                raise ValueError(f"Encountered unknown token '{text[i-1]}'.")

        loss = d[-1]
        return loss, p

    def backward_step(self, text, p):
        idx = len(p)
        tokenization = []
        while idx > 1:
            # move back the number of steps p tells you to
            next_idx = idx - p[idx - 1]

            # extract the final token
            tok = text[next_idx - 1 : idx - 1]
            tokenization.append(tok)

            idx = next_idx
        tokenization = list(reversed(tokenization))
        return tokenization

    def E_step(self, tokenization, trie):
        # get the new token counts based on updated tokenization
        counts = collections.Counter(tokenization)
        norm = sum(list(counts.values()))

        # Bayesianify them: https://cs.stanford.edu/~pliang/papers/tutorial-acl2007-talk.pdf
        # https://github.com/google/sentencepiece/blob/master/src/unigram_model_trainer.cc
        # we are returning the log probabilties here (alpha=0 prior)
        logsum = digamma(norm)
        for k, v in counts.items():
            counts[k] = digamma(v) - logsum

        for k, v in counts.items():
            trie.set_value(k, v)
        return trie

    def M_step(self, text, trie):
        loss, p = self.forward_step(text, trie)
        tokenization = self.backward_step(text, p)
        return tokenization, loss

    def EM_step(self, text, tokenization, trie):
        trie = self.E_step(tokenization, trie)
        tokenization, loss = self.M_step(text, trie)
        return loss, tokenization, trie

    def EM_round(self, text, tokens, delta=0.01, max_iter=10):
        tokenization, old_loss = self.M_step(text, self.trie)
        for step in range(max_iter):
            print(f"EM iter {step}: ", end="")
            loss, tokenization, trie = self.EM_step(text, tokenization, self.trie)
            print(f"Loss={loss:.2f}")
            if abs(old_loss - loss) < delta:
                break
            old_loss = loss

    def prune_tokens(self, tokens, characters, vocab_size, percent_to_prune=None):
        """Tokens are passed by reference and modified in place.
        Returns:
            True: to indicate to caller that more rounds are needed
            False: to indicate we successfully hit the target vocab size
            ValueError: if the vocab size cannot be reached."""
        sorted_tokens = tokens.most_common()
        N = len(sorted_tokens)
        if percent_to_prune is None:
            percent_to_prune = self.percent_to_prune
        n_trim = int(percent_to_prune * N)
        for i in reversed(range(N)):
            if N <= vocab_size:
                return False
            if n_trim <= 0:
                return True
            tok = sorted_tokens[i][0]
            if tok not in characters:
                self.trie.set_value(
                    tok, 0
                )  # we need to delete it from the trie (that sticks around)
                tokens.pop(
                    tok
                )  # also need to delete from tokens, so the next round doesn't see it
                n_trim -= 1
                N -= 1
        if n_trim > 0:
            raise ValueError(
                "Could not reduce tokens further. Please increase vocab size"
            )
        return False

    def fit(self, texts, vocab_size=None, delta=0.01, max_iter=5, max_rounds=5):
        """To turn off pruning, just set max_rounds=1"""
        # text = re.sub(" ", "_", text)
        text, tokens, characters = self.initialize_vocab(texts)
        if vocab_size is None:
            vocab_size = self.vocab_size

        if vocab_size > len(tokens):
            raise ValueError(
                f"Vocab size is larger than the availble number of tokens {len(tokens)}."
            )
        self.trie, self.maxlen = self._initialize_trie(tokens)
        for i in range(1, max_rounds + 1):
            print(f"--- Round {i}. Vocab size: {len(tokens)} ---")
            self.EM_round(text, tokens, delta, max_iter)
            if not self.prune_tokens(tokens, characters, vocab_size):
                break
        self.vocab_size = len(tokens)

    def generalized_forward_step(self, text, trie, nbest_size=1):
        N = len(text)
        d = [-np.inf] * (N + 1)
        p = [None] * (N + 1)
        d[0] = 0
        for i in range(1, N + 1):
            d_queue = []
            p_queue = []
            for j in range(max(i - self.maxlen, 0), i):
                final_token = text[j:i]
                final_value = trie.get_value(final_token)
                if final_value:
                    curr_d = d[j] + final_value
                    curr_p = len(final_token)
                    d[i] = max(d[i], curr_d)
                    d_queue.append(curr_d)
                    p_queue.append(curr_p)
            ids = np.argsort(d_queue)[-nbest_size:]
            p[i] = [p_queue[z] for z in ids]
        return p

    def generalized_backward_step(self, text, p):
        idx = len(p)
        tokenization = []
        while idx > 1:
            back_steps = np.random.choice(p[idx - 1])
            next_idx = idx - back_steps
            tok = text[next_idx - 1 : idx - 1]
            tokenization.append(tok)
            idx = next_idx
        tokenization = list(reversed(tokenization))
        return tokenization

    def tokenize(self, text, nbest_size=1):
        text = self.pre_tokenize(text)
        if self.trie is None:
            raise ValueError("Trainer has not yet been fit. Cannot tokenize.")
        p = self.generalized_forward_step(text, self.trie, nbest_size)
        tokenization = self.generalized_backward_step(text, p)
        return tokenization



import pubmed_parser


def parse_medline_xml(contents):
    try:
        return pubmed_parser.parse_medline_xml(contents)
    except Exception as e:
        print(e)
        return None


def parse_pubmed_paragraph(contents):
    try:
        return pubmed_parser.parse_pubmed_paragraph(contents, all_paragraph=False)
    except Exception as e:
        print(e)
        return None



# coding=utf-8
# Copyright 2018 The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert ELECTRA checkpoint."""


import os
import glob
import logging
import orjson as json
from pathlib import Path

import torch

from transformers import (
    ElectraConfig,
    ElectraForMaskedLM,
    ElectraForPreTraining,
    load_tf_weights_in_electra,
)

log = logging.getLogger(__name__)


def convert_tf_checkpoint_to_pytorch(
    tf_checkpoint_path, config_file, pytorch_dump_path, model_type, load_weights_func
):
    # Initialise PyTorch model
    config = ElectraConfig.from_json_file(config_file)
    print("Building PyTorch model from configuration: {}".format(str(config)))

    if model_type == "discriminator":
        model = ElectraForPreTraining(config)
    elif model_type == "generator":
        model = ElectraForMaskedLM(config)
    else:
        raise ValueError(
            "The discriminator_or_generator argument should be either 'discriminator' or 'generator'"
        )

    # Load weights from tf checkpoint
    globals()[load_weights_func](
        model, config, tf_checkpoint_path, discriminator_or_generator=model_type
    )

    # Save pytorch-model
    print("Save PyTorch model to {}".format(pytorch_dump_path))
    torch.save(model.state_dict(), pytorch_dump_path)


def load_tf2_weights_in_electra(
    model, config, tf_checkpoint_path, discriminator_or_generator="discriminator"
):
    """Load tf checkpoints in a pytorch model."""
    try:
        import re
        import numpy as np
        import tensorflow as tf
    except ImportError:
        log.error(
            "Loading a TensorFlow model in PyTorch, requires TensorFlow to be installed. Please see "
            "https://www.tensorflow.org/install/ for installation instructions."
        )
        raise
    tf_path = os.path.abspath(tf_checkpoint_path)
    log.info("Converting TensorFlow checkpoint from {}".format(tf_path))
    # Load weights from TF model
    init_vars = tf.train.list_variables(tf_path)
    names = []
    arrays = []
    log.info("Loading names and arrays from TensorFlow checkpoint.\n")
    for name, shape in init_vars:

        m_names = name.split("/")
        if name == "_CHECKPOINTABLE_OBJECT_GRAPH" or m_names[0] in [
            "global_step",
            "temperature",
            "save_counter",
            "phase",
            "step",
        ]:
            log.info(f" - Skipping non-model layer {name}")
            continue
        if "optimizer" in name:
            log.info(f" - Skipping optimization layer {name}")
            continue

        array = tf.train.load_variable(tf_path, name)
        arrays.append(array)

        # convet layer/num to layer_num
        m = re.search(r"layer\/\d+", name)
        if m:
            name = name.replace(m.group(0), m.group(0).replace("/", "_"))

        # remove variable name ending: .ATTRIBUTES/VARIABLE_VALUE
        name = name.replace("/.ATTRIBUTES/VARIABLE_VALUE", "")

        log.info("Loading TF weight {} with shape {}".format(name, shape))
        names.append(name)

    log.info("Converting TensorFlow weights to PyTorch weights.\n")
    for name, array in zip(names, arrays):
        original_name: str = name
        try:
            if discriminator_or_generator == "discriminator":
                if name.startswith("model/generator/") or name.startswith("generator/"):
                    log.info(f"Skipping generator: {original_name}")
                    continue
            elif discriminator_or_generator == "generator":
                if name.startswith("model/discriminator/") and not name.startswith(
                    "model/discriminator/electra/embeddings"
                ):
                    log.info(f"Skipping discriminator: {original_name}")
                    continue
                elif name.startswith("electra/") and not name.startswith(
                    "electra/embeddings/"
                ):
                    log.info(f"Skipping discriminator: {original_name}")
                    continue
                elif name.startswith("generator/"):
                    name = name.replace("generator/", "electra/")
            # in case, tensorflow version is higher than 2.0.0, change the names
            if name.startswith("model/discriminator/") or name.startswith(
                "model/generator/"
            ):
                name = name.replace("model/discriminator/", "")
                name = name.replace("model/generator/", "")
                name = name.replace("bert_output", "output")
                name = name.replace("dense_output", "output")
                name = name.replace("self_attention", "self")
                name = name.replace(
                    "/embeddings/position_embeddings/embeddings",
                    "/embeddings/position_embeddings",
                )
                name = name.replace(
                    "/embeddings/token_type_embeddings/embeddings",
                    "/embeddings/token_type_embeddings",
                )
                name = name.replace("/embeddings/weight", "/embeddings/word_embeddings")

            name = name.replace("dense_1", "dense_prediction")
            name = name.replace(
                "generator_predictions/output_bias", "generator_lm_head/bias"
            )

            name = name.split("/")
            # adam_v and adam_m are variables used in AdamWeightDecayOptimizer to calculated m and v
            pointer = model
            # model_names = dir(model)
            # logger.info(model_names)
            for m_name in name:
                if m_name not in [
                    "model",
                    ".OPTIMIZER_SLOT",
                    "optimizer",
                    "base_optimizer",
                    "v",
                    ".ATTRIBUTES",
                    "VARIABLE_VALUE",
                ]:
                    # logger.info(f'processing {m_name} from {name}')
                    if re.fullmatch(r"[A-Za-z]+_\d+", m_name):
                        scope_names = re.split(r"_(\d+)", m_name)
                    else:
                        scope_names = [m_name]
                    if scope_names[0] == "kernel" or scope_names[0] == "gamma":
                        pointer = getattr(pointer, "weight")
                    elif scope_names[0] == "output_bias" or scope_names[0] == "beta":
                        pointer = getattr(pointer, "bias")
                    elif scope_names[0] == "output_weights":
                        pointer = getattr(pointer, "weight")
                    elif scope_names[0] == "squad":
                        pointer = getattr(pointer, "classifier")
                    else:
                        pointer = getattr(pointer, scope_names[0])
                    if len(scope_names) >= 2:
                        num = int(scope_names[1])
                        pointer = pointer[num]
            if m_name.endswith("_embeddings"):
                pointer = getattr(pointer, "weight")
            elif m_name == "kernel":
                array = np.transpose(array)
            try:
                assert pointer.shape == array.shape, original_name
            except AssertionError as e:
                e.args += (pointer.shape, array.shape)
                raise
            print("Initialize PyTorch weight {}".format(name), original_name)
            pointer.data = torch.from_numpy(array)
        except AttributeError as e:
            print("Skipping {}".format(original_name), name, e)
            continue
    return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Korean Sentence Splitter
# Split Korean text into sentences using heuristic algorithm.
#
# Copyright (C) 2021 Hyun-Woong Ko <kevin.ko@tunib.ai> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import logging
from typing import List

from .character_definition import (
    get_emoji,
    categories,
)
from .rule import Table, Stats


class Const:
    numbers = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"]
    lower_alphabets = [
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
    ]
    upper_alphabets = [_.upper() for _ in lower_alphabets]

    # <와 >는 수학기호나 화살표(<-, ->)로 쓰이는 경우가 잦아서 제외 함
    brackets = [
        ")",
        "）",
        "〉",
        "》",
        "]",
        "］",
        "〕",
        "】",
        "}",
        "｝",
        "』",
        "」",
        "(",
        "（",
        "〈",
        "《",
        "[",
        "［",
        "〔",
        "【",
        "{",
        "｛",
        "「",
        "『",
    ]

    bracket_open_to_close = {
        "(": ")",
        "（": "）",
        "〈": "〉",
        "《": "》",
        "[": "]",
        "［": "］",
        "〔": "〕",
        "【": "】",
        "{": "}",
        "｛": "｝",
        "「": "」",
        "『": "』",
    }

    bracket_close_to_open = {v: k for k, v in bracket_open_to_close.items()}

    double_quotes = ['"', "“", "”", "″"]
    single_quotes = ["'", "‘", "’", "`"]
    double_quotes_open_to_close = {
        "“": "”",
        '"': '"',
        "″": "″",
    }
    single_quotes_open_to_close = {
        "‘": "’",
        "'": "'",
        "`": "`",
    }
    double_quotes_close_to_open = {v: k for k, v in double_quotes_open_to_close.items()}
    single_quotes_close_to_open = {v: k for k, v in single_quotes_open_to_close.items()}

    punctuation = [";", ".", "?", "!", "~", "…"]
    special = punctuation + brackets
    quotes_or_brackets = single_quotes + double_quotes + brackets
    endpoint = (
        quotes_or_brackets + punctuation + [" "] + list(Table[Stats.COMMON].keys())
    )
    not_endpoint = [",", ":", "\u200b"] + quotes_or_brackets + ["<", ">"]

    @staticmethod
    def exceptions():
        faces = [":)", ":(", ":'(", "O:)", "&)", ">:(", "3:)", '<(")', ":-)", ":-("]
        low_upper = Const.lower_alphabets + Const.upper_alphabets

        apostrophe = []
        for i in low_upper:
            for j in low_upper:
                for k in Const.single_quotes:
                    apostrophe.append(f"{i}{k}{j}")

        year_s = []
        for i in Const.numbers:
            for j in Const.single_quotes:
                year_s.append(f"{i}{j}s")
                year_s.append(f"{i}{j}S")

        time = []
        for i in Const.numbers:
            for j in Const.numbers:
                for k in Const.single_quotes:
                    time.append(f"{i}{k}{j}")
                    time.append(f"{i}{k}{j}")
                    time.append(f"{i}{k}{j}")

        inch = []
        for i in Const.numbers + ["."]:
            for j in Const.numbers:
                for k in Const.double_quotes:
                    inch.append(f"{i}{j}{k}")

        return faces + apostrophe + year_s + time + inch

    @staticmethod
    def ec_cases():
        return [
            "쌓이다",
            "보이다",
            "먹이다",
            "죽이다",
            "끼이다",
            "트이다",
            "까이다",
            "꼬이다",
            "데이다",
            "치이다",
            "쬐이다",
            "꺾이다",
            "낚이다",
            "녹이다",
            "벌이다",
            "다 적발",
            "다 말하",
            "다 말한",
            "다 말했",
            "다 밝혀",
            "다 밝혔",
            "다 밝히",
            "다 밝힌",
            "다 주장",
            "요 라고",
            "요. 라고",
            "죠 라고",
            "죠. 라고",
            "다 라고",
            "다. 라고",
            "다 하여",
            "다 거나",
            "다. 거나",
            "다 시피",
            "다. 시피",
            "다 응답",
            "다 로 응답",
            "다. 로 응답",
            "요 로 응답",
            "요. 로 응답",
            "죠 로 응답",
            "죠. 로 응답",
            "다 에서",
            "다. 에서",
            "요 에서",
            "요. 에서",
            "죠 에서",
            "죠. 에서",
            "타다 금지법",
            "다 온 사실",
            "다 온 것",
            "다 온 사람",
            "다 왔다",
            "다 왔더",
            "다 와보",
            "우간다",
            "사이다",
        ]


class Eojeol:
    eojeol: str
    pos: str

    def __init__(self, eojeol: str = "", pos: str = "TEMP"):
        self.eojeol = eojeol
        self.pos = pos

    def __str__(self):
        return f"({self.eojeol}, {self.pos})"

    def __repr__(self):
        return f"('{self.eojeol}', {self.pos})"


class Preprocessor:
    def __init__(self, use_morpheme: bool):
        # global _exceptions
        # exceptions = deepcopy(_exceptions)
        exceptions = Const.exceptions()

        # drop ec cases if morpheme features are not available
        if not use_morpheme:
            exceptions += Const.ec_cases()

        self.backup_dict = {k: str(abs(hash(k))) for k in exceptions}

    @staticmethod
    def tostring(eojeols):
        return ["".join([j.eojeol for j in i]) for i in eojeols]

    @staticmethod
    def _replace(text: str, purpose_dict: dict):
        for k, v in purpose_dict.items():
            text = text.replace(k, v)

        return text

    def backup(self, text: str):
        return self._replace(text=text, purpose_dict=self.backup_dict)

    def restore(self, text: str):
        return self._replace(
            text=text,
            purpose_dict={v: k for k, v in self.backup_dict.items()},
        )

    def _add_item_to_dict(self, key: str, val: str):
        self.backup_dict[key] = val

    def add_ec_cases_to_dict(self, text):
        for i in range(0, len(text)):
            cond1 = text[i] in ["다", "요", "죠"]
            cond2 = i != len(text) - 1
            if cond1 and cond2:
                if text[i + 1] not in Const.endpoint:
                    target_to_backup = text[i] + text[i + 1]
                    self._add_item_to_dict(
                        key=target_to_backup,
                        val=str(abs(hash(target_to_backup))),
                    )

        return text

    def add_emojis_to_dict(self, text):
        for e in get_emoji(text):
            self._add_item_to_dict(
                key=e,
                val=str(abs(hash(e))),
            )
        return text


class Postprocessor(object):
    @staticmethod
    def _contains(src, tgt):
        for s in src:
            for p in [" ", "?", ".", "!"]:
                if f"{s}{p}" in tgt:
                    return True
        return False

    @staticmethod
    def _lindex_split(some_list, *args):
        args = (0,) + tuple(data + 1 for data in args) + (len(some_list) + 1,)
        return [some_list[start:end] for start, end in zip(args, args[1:])]

    @staticmethod
    def _find_all(a_str, sub):
        start = 0
        while True:
            start = a_str.find(sub, start)
            if start == -1:
                return
            yield start + len(sub) - 1
            start += len(sub)

    def _heuristic(self, results, post_processing_list):
        final_results = []
        for res in results:
            split_idx = []

            find_quotes = False
            for qt in Const.quotes_or_brackets:
                if qt in res:
                    find_quotes = True
                    break

            if not find_quotes:
                for post in post_processing_list:
                    if post in res:
                        split_idx += list(self._find_all(res, post))

            split_idx = sorted(split_idx)
            final_results += self._lindex_split(res, *split_idx)

        return final_results

    def apply_heuristic(self, text, results, use_morpheme):
        from kss.rule import (
            post_processing_da,
            post_processing_jyo,
            post_processing_yo,
        )

        if self._contains(["요"], text):
            results = self._heuristic(results, post_processing_yo)
        if self._contains(["죠"], text):
            results = self._heuristic(results, post_processing_jyo)
        if not use_morpheme and self._contains(["다"], text):
            results = self._heuristic(results, post_processing_da)

        return results


def empty(obj, dim=1) -> bool:
    assert dim in [1, 2], "only 1 or 2 dimension iterable is supported."

    if dim == 1:
        return len(obj) == 0
    else:
        return all([empty(o) for o in obj])


def top(stack: List[str], symbol: str) -> bool:
    return stack[len(stack) - 1] == symbol


def do_push_pop_symbol(stack: List[str], symbol: str, current_ch: str):
    # call by assignment
    if empty(stack):
        stack.append(symbol)

    else:
        if top(stack, current_ch):
            return stack.pop()
        else:
            stack.append(symbol)

    return current_ch


def check_pos(pos, pos_list):
    for target in pos_list:
        if target in pos.pos:
            return True
    return False


def replace_quotes_or_brackets_to_zwsp(eojeols):
    results = []
    for eojeol in eojeols:
        item = [eojeol]

        for s in Const.quotes_or_brackets:
            if eojeol.eojeol == s:
                item = [
                    Eojeol("\u200b", "TEMP"),
                    eojeol,
                    Eojeol("\u200b", "TEMP"),
                ]

        results += item

    return results


def length_constraints(
    text,
    max_recover_length,
    max_recover_step,
):
    if len(text) > max_recover_length and max_recover_step > 0:
        logging.warning(
            f"Too long text! turn off quotes calibration! text len:{len(text)}"
        )
        max_recover_step = 0

    return max_recover_step


def build_preprocessed_list(text):
    input_texts = []
    split_texts = text.split("\n")

    for t in split_texts:
        t = t.split("\t")
        for s in t:
            s = s.strip()
            if len(s) > 0:
                s = preprocess_text(s)
                input_texts.append(s)

    return input_texts


def clear_list_to_sentences(results):
    total_result = []
    for result in results:
        temp = []
        for sent in result:
            sent = sent.strip()
            if len(sent) > 0:
                temp.append(sent)
        total_result.append(temp)

    return total_result


def get_num_workers(num_workers):
    if num_workers < 0:
        num_workers = None

    if num_workers == 0:
        num_workers = 1

    return num_workers


def remove_useless_space(text):
    text = text.replace("\n", "")
    while "  " in text:
        text = text.replace("  ", " ")

    return text


def get_chunk_with_index(text, span):
    text = remove_useless_space(text)
    start = span[0][0]
    end = span[-1][1]
    return text[start:end]


def preprocess_text(text):
    total_text = ""
    for t in text:
        if t in _posix:
            total_text += t
        elif len(get_emoji(t)) != 0:
            total_text += t

    while "  " in total_text:
        total_text = total_text.replace("  ", " ")

    return total_text


_posix = list(chr(x) for x in categories.keys())
# _exceptions = Const.exceptions()



from typing import Any, Callable, Dict, List, NamedTuple, Optional

import numpy as np
import sklearn.metrics as skmetrics

from ..utils import filter_labels, to_int_label_array


class Metric(NamedTuple):
    """Specification for a metric and the subset of [golds, preds, probs] it expects."""

    func: Callable[..., float]
    inputs: List[str] = ["golds", "preds"]


def metric_score(
    golds: Optional[np.ndarray] = None,
    preds: Optional[np.ndarray] = None,
    probs: Optional[np.ndarray] = None,
    metric: str = "accuracy",
    filter_dict: Optional[Dict[str, List[int]]] = None,
    **kwargs: Any,
) -> float:
    """Evaluate a standard metric on a set of predictions/probabilities.

    Parameters
    ----------
    golds
        An array of gold (int) labels
    preds
        An array of (int) predictions
    probs
        An [n_datapoints, n_classes] array of probabilistic (float) predictions
    metric
        The name of the metric to calculate
    filter_dict
        A mapping from label set name to the labels that should be filtered out for
        that label set

    Returns
    -------
    float
        The value of the requested metric

    Raises
    ------
    ValueError
        The requested metric is not currently supported
    ValueError
        The user attempted to calculate roc_auc score for a non-binary problem
    """
    if metric not in METRICS:
        msg = f"The metric you provided ({metric}) is not currently implemented."
        raise ValueError(msg)

    # Print helpful error messages if golds or preds has invalid shape or type
    golds = to_int_label_array(golds) if golds is not None else None
    preds = to_int_label_array(preds) if preds is not None else None

    # Optionally filter out examples (e.g., abstain predictions or unknown labels)
    label_dict: Dict[str, Optional[np.ndarray]] = {
        "golds": golds,
        "preds": preds,
        "probs": probs,
    }
    if filter_dict:
        if set(filter_dict.keys()).difference(set(label_dict.keys())):
            raise ValueError(
                "filter_dict must only include keys in ['golds', 'preds', 'probs']"
            )
        # label_dict is overwritten from type Dict[str, Optional[np.ndarray]]
        # to Dict[str, np.ndarray]
        label_dict = filter_labels(label_dict, filter_dict)  # type: ignore

    # Confirm that required label sets are available
    func, label_names = METRICS[metric]
    for label_name in label_names:
        if label_dict[label_name] is None:
            raise ValueError(f"Metric {metric} requires access to {label_name}.")

    label_sets = [label_dict[label_name] for label_name in label_names]
    return func(*label_sets, **kwargs)


def _coverage_score(preds: np.ndarray) -> float:
    return np.sum(preds != -1) / len(preds)


def _roc_auc_score(golds: np.ndarray, probs: np.ndarray) -> float:
    if not probs.shape[1] == 2:
        raise ValueError(
            "Metric roc_auc is currently only defined for binary problems."
        )
    return skmetrics.roc_auc_score(golds, probs[:, 1])


def _f1_score(golds: np.ndarray, preds: np.ndarray) -> float:
    if golds.max() <= 1:
        return skmetrics.f1_score(golds, preds)
    else:
        raise ValueError(
            "f1 not supported for multiclass. Try f1_micro or f1_macro instead."
        )


def _f1_micro_score(golds: np.ndarray, preds: np.ndarray) -> float:
    return skmetrics.f1_score(golds, preds, average="micro")


def _f1_macro_score(golds: np.ndarray, preds: np.ndarray) -> float:
    return skmetrics.f1_score(golds, preds, average="macro")


# See https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics
# for details on the definitions and available kwargs for all metrics from scikit-learn
METRICS = {
    "accuracy": Metric(skmetrics.accuracy_score),
    "coverage": Metric(_coverage_score, ["preds"]),
    "precision": Metric(skmetrics.precision_score),
    "recall": Metric(skmetrics.recall_score),
    "f1": Metric(_f1_score, ["golds", "preds"]),
    "f1_micro": Metric(_f1_micro_score, ["golds", "preds"]),
    "f1_macro": Metric(_f1_macro_score, ["golds", "preds"]),
    "fbeta": Metric(skmetrics.fbeta_score),
    "matthews_corrcoef": Metric(skmetrics.matthews_corrcoef),
    "roc_auc": Metric(_roc_auc_score, ["golds", "probs"]),
}
